from dexy.utils import getdoc
from dexy.utils import parse_json
from dexy.utils import parse_yaml
from dexy.utils import s
from dexy.version import DEXY_VERSION
from modargs import args
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonLexer
import dexy.exceptions
import dexy.plugins # so all built-in plugins are registered
import dexy.wrapper
import inspect
import json
import logging
import os
import sys
import warnings
import yaml

D = dexy.wrapper.Wrapper.DEFAULTS

DEFAULT_COMMAND = 'dexy'
MOD = sys.modules[__name__]
PROG = 'dexy'
S = "   "

def parse_and_run_cmd(argv, module, default_command):
    try:
        args.parse_and_run_command(argv, module, default_command)
    except dexy.exceptions.UserFeedback as e:
        sys.stderr.write("Oops, there's a problem running your command. Here is the error message:" + os.linesep)
        err_msg = str(e)
        if err_msg:
            sys.stderr.write("'%s'" % str(e))
        else:
            sys.stderr.write("Sorry, can't get text of error message.")
        sys.stderr.write(os.linesep)
        sys.exit(1)
    except KeyboardInterrupt:
        sys.stderr.write("""
        ok, stopping your dexy run
        you might want to 'dexy reset' before running again\n""")
        sys.exit(1)

def run():
    """
    Method that runs the command specified on the command line.
    """
    if hasattr(logging, 'captureWarnings'):
        logging.captureWarnings(True)
    else:
        warnings.filterwarnings("ignore",category=Warning)

    if len(sys.argv) == 1 or (sys.argv[1] in args.available_commands(MOD)) or sys.argv[1].startswith("-"):
        parse_and_run_cmd(sys.argv[1:], MOD, default_command=DEFAULT_COMMAND)

    else:
        if ":" in sys.argv[1]:
            command, subcommand = sys.argv[1].split(":")
        else:
            command = sys.argv[1]
            subcommand = ""

        command_class = dexy.plugin.Command.aliases.get(command)

        if not command_class:
            args.parse_and_run_command(subcommand, dexy.commands)

        mod_name = command_class.__module__
        mod = args.load_module(mod_name)

        if command_class.DEFAULT_COMMAND:
            default_command = command_class.DEFAULT_COMMAND
        else:
            default_command = command_class.NAMESPACE

        parse_and_run_cmd([subcommand] + sys.argv[2:], mod, default_command=default_command)

def config_args(modargs):
    cliargs = modargs.get("__cli_options", {})
    kwargs = modargs.copy()

    config_file = modargs.get('conf', dexy.wrapper.Wrapper.DEFAULTS['config_file'])

    # Update from config file
    if os.path.exists(config_file):
        with open(config_file, "rb") as f:
            if config_file.endswith(".conf"):
                try:
                    conf_args = parse_yaml(f.read())
                except dexy.exceptions.UserFeedback as yaml_exception:
                    try:
                        conf_args = parse_json(f.read())
                    except dexy.exceptions.UserFeedback as json_exception:
                        print "--------------------------------------------------"
                        print "Tried to parse YAML:"
                        print yaml_exception
                        print "--------------------------------------------------"
                        print "Tried to parse JSON:"
                        print json_exception
                        print "--------------------------------------------------"
                        raise dexy.exceptions.UserFeedback("Unable to parse config file '%s' as YAML or as JSON." % config_file)

            elif config_file.endswith(".yaml"):
                conf_args = parse_yaml(f.read())
            elif config_file.endswith(".json"):
                conf_args = parse_json(f.read())
            else:
                raise dexy.exceptions.UserFeedback("Don't know how to load config from '%s'" % config_file)

            kwargs.update(conf_args)

    if cliargs: # cliargs may be False
        for k in cliargs.keys(): kwargs[k] = modargs[k]

    # TODO allow updating from env variables, e.g. DEXY_ARTIFACTS_DIR

    return kwargs

RENAME_PARAMS = {
        'artifactsdir' : 'artifacts_dir',
        'conf' : 'config_file',
        'dbalias' : 'db_alias',
        'dbfile' : 'db_file',
        'disabletests' : 'disable_tests',
        'dryrun' : 'dry_run',
        'excludealso' : 'exclude_also',
        'ignore' : 'ignore_nonzero_exit',
        'logfile' : 'log_file',
        'logformat' : 'log_format',
        'loglevel' : 'log_level',
        'logdir' : 'log_dir',
        'nocache' : 'dont_use_cache'
        }

def rename_params(kwargs):
    renamed_args = {}
    for k, v in kwargs.iteritems():
        renamed_key = RENAME_PARAMS.get(k, k)
        renamed_args[renamed_key] = v
    return renamed_args

def skip_params(kwargs):
    ok_params = {}
    for k, v in kwargs.iteritems():
        if k in D.keys():
            ok_params[k] = v
    return ok_params

def init_wrapper(modargs):
    kwargs = config_args(modargs)
    kwargs = rename_params(kwargs)
    kwargs = skip_params(kwargs)
    return dexy.wrapper.Wrapper(**kwargs)

def default_config():
    wrapper = dexy.wrapper.Wrapper()
    conf = wrapper.__dict__.copy()

    for k in conf.keys():
        if not k in D.keys():
            del conf[k]

    reverse_rename = dict((v,k) for k, v in RENAME_PARAMS.iteritems())
    for k in conf.keys():
        renamed_key = reverse_rename.get(k, k)
        if renamed_key != k:
            conf[renamed_key] = conf[k]
            del conf[k]

    return conf

def run_dexy_in_profiler(wrapper, profile):
    if isinstance(profile, bool):
        profile_filename = 'dexy.prof'
    else:
        profile_filename = profile

    import cProfile
    print "running dexy with cProfile, writing profile data to %s" % profile_filename
    cProfile.runctx("wrapper.run()", None, locals(), profile_filename)
    import pstats
    stat = pstats.Stats(profile_filename)
    stat.sort_stats("cumulative")
    stat.print_stats(25)

def handle_user_feedback_exception(wrapper, e):
    if hasattr(wrapper, 'log'):
        wrapper.log.error("A problem has occurred with one of your documents:")
        wrapper.log.error(e.message)
    wrapper.cleanup_partial_run()
    sys.stderr.write("Oops, there's a problem processing one of your documents. Here is the error message:" + os.linesep)
    sys.stderr.write(e.message)
    if not e.message.endswith(os.linesep) or e.message.endswith("\n"):
        sys.stderr.write(os.linesep)
    sys.stderr.write("Dexy is stopping. There might be more information at the end of the dexy log." + os.linesep)
    sys.exit(1)

def log_and_print_exception(wrapper, e):
    if hasattr(wrapper, 'log'):
        wrapper.log.error("An error has occurred.")
        wrapper.log.error(e)
        wrapper.log.error(e.message)
    import traceback
    traceback.print_exc()

def dexy_command(
        __cli_options=False,
        artifactsdir=D['artifacts_dir'], # location of directory in which to store artifacts
        conf=D['config_file'], # name to use for configuration file
        danger=D['danger'], # whether to allow running remote files
        dbalias=D['db_alias'], # type of database to use
        dbfile=D['db_file'], # name of the database file (it lives in the logs dir)
        disabletests=D['disable_tests'], # Whether to disable the dexy 'test' filter
        dryrun=D['dry_run'], # if True, just parse config and print batch info, don't run dexy
        encoding=D['encoding'], # Default encoding. Set to 'chardet' to use chardet auto detection.
        exclude=D['exclude'], # comma-separated list of directory names to exclude from dexy processing
        excludealso=D['exclude_also'], # comma-separated list of directory names to exclude from dexy processing
        full=D['full'], # Whether to do a full run including tasks marked default: False
        globals=D['globals'], # global values to make available within dexy documents, should be KEY=VALUE pairs separated by spaces
        help=False, # for people who type -help out of habit
        h=False, # for people who type -h out of habit
        hashfunction=D['hashfunction'], # What hash function to use, set to crc32 or adler32 for more speed but less reliability
        ignore=D['ignore_nonzero_exit'], # whether to ignore nonzero exit status or raise an error - may not be supported by all filters
        logfile=D['log_file'], # name of log file
        logformat=D['log_format'], # format of log entries
        loglevel=D['log_level'], # log level, valid options are DEBUG, INFO, WARN
        logdir=D['log_dir'], # location of directory in which to store logs
        nocache=D['dont_use_cache'], # whether to force artifacts to run even if there is a matching file in the cache
        profile=D['profile'], # whether to run with cProfile. Arg can be a boolean, in which case profile saved to 'dexy.prof', or a filename to save to.
        recurse=D['recurse'], # whether to recurse into subdirectories when running Dexy
        r=False, # whether to clear cache before running dexy
        reset=False, # whether to clear cache before running dexy
        reports=D['reports'], # reports to be run after dexy runs, enclose in quotes and separate with spaces
        siblings=D['siblings'], # whether siblings should have prior siblings as inputs (slows dexy down on large projects, siblings should run in order regardless)
        silent=D['silent'], # Whether to not print any output when running dexy
        uselocals=D['uselocals'], # use cached local copies of remote URLs, faster but might not be up to date, 304 from server will override this setting
        target=D['target'], # Which target to run. By default all targets are run, this allows you to run only 1 bundle (and its dependencies).
        version=False # For people who type -version out of habit
    ):
    """
    Runs Dexy.
    """
    if h or help:
        return help_command()

    if version:
        return version_command()

    if r or reset:
        print "Resetting dexy cache..."
        reset_command(artifactsdir=artifactsdir, logdir=logdir)

    try:
        wrapper = init_wrapper(locals())

        if profile:
            run_dexy_in_profiler(wrapper, profile)
        else:
            wrapper.run()

        wrapper.report()
        print "finished in %0.4f" % wrapper.batch.elapsed()

    except dexy.exceptions.UserFeedback as e:
        handle_user_feedback_exception(wrapper, e)
    except Exception as e:
        log_and_print_exception(wrapper, e)

def reset_command(
        __cli_options=False,
        artifactsdir=D['artifacts_dir'], # location of directory in which to store artifacts
        logdir=D['log_dir']# location of directory in which to store logs
        ):
    """
    Empty the artifacts and logs directories.
    """
    wrapper = init_wrapper(locals())
    wrapper.remove_dexy_dirs()
    wrapper.setup_dexy_dirs()

def cleanup_command(
        __cli_options=False,
        artifactsdir=D['artifacts_dir'], # location of directory in which to store artifacts
        logdir=D['log_dir'], # location of directory in which to store logs
        reports=True # Also remove report generated dirs
        ):
    """
    Remove the artifacts and logs directories.
    """
    wrapper = init_wrapper(locals())
    wrapper.remove_dexy_dirs(reports)

def setup_command(__cli_options=False, **kwargs):
    """
    Create the directories dexy needs to run. This helps make sure you mean to run dexy in this directory.
    """
    wrapper = init_wrapper(locals())
    wrapper.setup_dexy_dirs()

def reports_command():
    def sort_key(k):
        return k.__name__

    report_classes = sorted(dexy.reporter.Reporter.plugins, key=sort_key)
    for klass in report_classes:
        print "%s: %s" % (klass.__name__, ", ".join(klass.ALIASES))

def help_command(
        example=False, # Whether to run any live examples, if available.
        filters=False, # Whether to print the list of available dexy filters.
        reports=False, # Whether to print the list of available dexy reports.
        f=False, # If a filter alias is specified, help for that filter is printed.
        on=False # The dexy command to get help on.
        ):

    if f:
        filter_command(f, example)
    elif filters:
        filters_command()
    elif reports:
        reports_command()
    else:
        args.help_command(PROG, MOD, DEFAULT_COMMAND, on)

def help_text(on=False):
    return args.help_text(PROG, MOD, DEFAULT_COMMAND, on)

def version_command():
    """Print the current version."""
    print "%s version %s" % (PROG, DEXY_VERSION)

def conf_command(
        conf=D['config_file'], # name of config file to write to
        p=False # whether to print to stdout rather than write to file
        ):
    """
    Write a config file containing dexy's defaults.
    """
    if os.path.exists(conf) and not p:
        print "Config file %s already exists, will print conf to stdout instead..." % conf
        p = True

    config = default_config()

    # No point specifying config file name in config file.
    del config['conf']

    YAML_HELP = """# YAML config file for dexy.
# You can delete any lines you don't wish to customize.
# Options are same as command line options, for more info run 'dexy help -on dexy'.\n"""

    if p:
        print yaml.dump(config, default_flow_style=False)
    else:
        with open(conf, "wb") as f:
            if conf.endswith(".yaml") or conf.endswith(".conf"):
    
                f.write(YAML_HELP)
                f.write(yaml.dump(config, default_flow_style=False))
            elif conf.endswith(".json"):
                json.dump(config, f, sort_keys=True, indent=4)
            else:
                raise dexy.exceptions.UserFeedback("Don't know how to write config file '%s'" % conf)

        print "Config file has been written to '%s'" % conf

def filter_command(
        alias="", # If a filter alias is specified, more detailed help for that filter is printed.
        example=False, # Whether to run examples
        nocolor=False, # When source = True, whether to omit syntax highlighting
        showall=False, # Whether to show all filters, including those which need missing software, implies versions=True
        showmissing=False, # Whether to just show filters missing external software, implies versions=True
        space=False, # Whether to add extra spacing to the output for extra readability
        source=False, # Whether to include syntax-highlighted source code when displaying an indvidual filter
        versions=False # Whether to check the installed version of external software required by filters, slower
        ):
    """
    Information about available dexy filters.
    """
    print filters_text(**locals())

def filters_command(
        alias="", # If a filter alias is specified, more detailed help for that filter is printed.
        example=False, # Whether to run examples
        nocolor=False, # When source = True, whether to omit syntax highlighting
        showall=False, # Whether to show all filters, including those which need missing software, implies versions=True
        showmissing=False, # Whether to just show filters missing external software, implies versions=True
        space=False, # Whether to add extra spacing to the output for extra readability
        source=False, # Whether to include syntax-highlighted source code when displaying an indvidual filter
        versions=False # Whether to check the installed version of external software required by filters, slower
        ):
    """
    Information about available dexy filters.
    """
    print filters_text(**locals())

def filters_text(
        alias="", # If a filter alias is specified, more detailed help for that filter is printed.
        example=False, # Whether to run examples
        nocolor=False, # When source = True, whether to omit syntax highlighting
        showall=False, # Whether to show all filters, including those which need missing software, implies versions=True
        showmissing=False, # Whether to just show filters missing external software, implies versions=True
        space=False, # Whether to add extra spacing to the output for extra readability
        source=False, # Whether to include syntax-highlighted source code when displaying an indvidual filter
        versions=False # Whether to check the installed version of external software required by filters, slower
        ):

    if len(alias) > 0:
        # We want help on a particular filter
        klass = dexy.filter.Filter.aliases[alias]
        text = []
        text.append("aliases: %s" % ", ".join(klass.ALIASES))
        text.append("")
        text.append(inspect.getdoc(klass))
        text.append("")
        text.append("input formats: %s" % ", ".join(klass.INPUT_EXTENSIONS))
        text.append("output formats: %s" % ", ".join(klass.OUTPUT_EXTENSIONS))

        templates = klass.templates()
        if len(templates) > 0:
            text.append("")
            text.append("Templates which use this filter:")
            for t in templates:
                aliases = [k for k, v in dexy.template.Template.aliases.iteritems() if v == t]
                text.append("")
                text.append("  %s" % aliases[0])
                text.append("            %s" % dexy.utils.getdoc(t))

            if example:
                for t in templates:
                    aliases = [k for k, v in dexy.template.Template.aliases.iteritems() if v == t]
                    if t.__module__ == "dexy_filter_examples":
                        text.append('')
                        text.append("Running example: %s" % s(t.__doc__))
                        text.append('')
                        text.append('')
                        text.append(template_text(alias=aliases[0]))
                        text.append('')
        text.append("")
        text.append("For online docs see http://dexy.it/docs/filters/%s" % alias)
        if source:
            text.append("")
            source_code = inspect.getsource(klass)
            if nocolor:
                text.append(source_code)
            else:
                formatter = TerminalFormatter()
                lexer = PythonLexer()
                text.append(highlight(source_code, lexer, formatter))
        return "\n".join(text)

    else:
        def sort_key(k):
            if len(k.ALIASES) == 0:
                return None
            else:
                return k.ALIASES[0]

        filter_classes = sorted(set(f for f in dexy.filter.Filter.plugins), key=sort_key)

        text = []

        text.append("Available filters:")
        for klass in filter_classes:
            if not showall:
                skip = (len(klass.ALIASES) == 0) or klass.NODOC
            else:
                skip = False

            if (versions or showmissing or showall) and not skip:
                if hasattr(klass, 'version'):
                    version = klass.version()
                else:
                    version = None
                no_version_info_available = (version is None)
                if no_version_info_available:
                    version_message = ""
                    if showmissing:
                        skip = True
                elif version:
                    version_message = "Installed version: %s" % version
                    if showmissing:
                        skip = True
                else:
                    if not (showmissing or showall):
                        skip = True
                    version_message = "'%s' failed, filter may not be available." % klass.version_command()

            if not skip:
                aliases = ", ".join(klass.ALIASES)
                filter_help = "  " + aliases + " : " + getdoc(klass)
                if (versions or showmissing or (showall and not version)):
                    filter_help += " %s" % version_message
                text.append(filter_help)

        text.append("\nFor more information about a particular filter, use the -alias flag and specify the filter alias.")
        if space:
            sep = "\n\n"
        else:
            sep = "\n"
        return sep.join(text)

def it_command(**kwargs):
    # so you can type 'dexy it' if you want to
    dexy_command(kwargs)

def grep_command(
        __cli_options=False, # nodoc
        expr=None, # The expression to search for
        keyexpr="", # Only search for keys matching this expression, implies keys=True
        keys=False, # if True, try to list the keys in any found files
        recurse=False, # if True, recurse into keys to look for sub keys (implies keys=True)
        artifactsdir=D['artifacts_dir'], # location of directory in which to store artifacts
        logdir=D['log_dir'] # location of directory in which to store logs
        ):
    """
    Search for a Dexy document in the database matching the expression.

    For sqlite the expression will be wrapped in % for you.
    """
    wrapper = init_wrapper(locals())
    wrapper.setup_read()

    for row in wrapper.db.query_docs("%%%s%%" % expr):
        print row['key']
#        if keys or len(keyexpr) > 0 or recurse:
#            artifact_classes = dexy.introspect.artifact_classes()
#            artifact_class = artifact_classes[artifactclass]
#            artifact = artifact_class.retrieve(row['hashstring'])
#            if artifact.ext in [".json", ".kch", ".sqlite3"]:
#                if len(keyexpr) > 0:
#                    rows = artifact.kv_storage().query("%%%s%%" % keyexpr)
#                else:
#                    rows = artifact.kv_storage().keys()
#
#                if rows:
#                    print "  key-value store keys:"
#                for k in rows:
#                    print "    %s" % k
#                    if recurse:
#                        v = artifact.retrieve_from_kv_storage(k)
#                        try:
#                            if not hasattr(v, "keys"):
#                                v = json.loads(v)
#                            if hasattr(v, "keys"):
#                                for kk in v.keys():
#                                    print "      %s" % kk
#                        except Exception as e:
#                            pass
#
#            if len(artifact.data_dict.keys()) > 1:
#                print "  data dict keys:"
#            for k in artifact.data_dict.keys():
#                if not k == '1':
#                    print "    %s" % k

def fcmds_command(alias=False):
    """
    Returns a list of available filter commands (fcmds) defined by the specified alias.

    These commands can then be run using the fcmd command.
    """

    def filter_class_commands(filter_alias):
        filter_class = dexy.filter.Filter.aliases[filter_alias]
        cmds = []
        for m in dir(filter_class):
            if m.startswith("docmd_"):
                cmds.append(m.replace("docmd_", ""))
        return sorted(cmds)

    filters_dict = dexy.filter.Filter.aliases
    if (not alias) or (not alias in filters_dict):
        print "Aliases with filter commands defined are:"
        for a in sorted(filters_dict):
            cmds = filter_class_commands(a)
            if len(cmds) > 0:
                print a
    else:
        print "Filter commands defined for %s:" % alias
        cmds = filter_class_commands(alias)
        print os.linesep.join(cmds)

def fcmd_command(
        alias=None, # The alias of the filter which defines the custom command
        cmd=None, # The name of the command to run
        help=False, # If true, just print docstring rather than running command
        **kwargs # Additional arguments to be passed to the command
        ):
    """
    Run a command defined in a dexy filter.
    """
    filter_class = dexy.filter.Filter.aliases.get(alias)

    if not filter_class:
        raise dexy.exceptions.UserFeedback("%s is not a valid alias" % alias)

    cmd_name = "docmd_%s" % cmd

    if not filter_class.__dict__.has_key(cmd_name):
        raise dexy.exceptions.UserFeedback("%s is not a valid command. There is no method %s defined in %s" % (cmd, cmd_name, filter_class.__name__))
    else:
        class_method = filter_class.__dict__[cmd_name]
        if type(class_method) == classmethod:
            if help:
                print inspect.getdoc(class_method.__func__)
            else:
                try:
                    class_method.__func__(filter_class, **kwargs)
                except TypeError as e:
                    print e.message
                    print inspect.getargspec(class_method.__func__)
                    print inspect.getdoc(class_method.__func__)
                    raise e

        else:
            raise dexy.exceptions.InternalDexyProblem("expected %s to be a classmethod of %s" % (cmd_name, filter_class.__name__))

def reporters_command(
        ):
    """
    List available reporters.
    """
    FMT = "%-10s %-20s %s"

    if dexy.reporter.Reporter.aliases:
        print FMT % ('ALIAS', 'DIRECTORY', 'INFO')

    for k in sorted(dexy.reporter.Reporter.aliases):
        reporter_class = dexy.reporter.Reporter.aliases[k]
        reports_dir = reporter_class.REPORTS_DIR or ''
        print FMT % (k, reports_dir, getdoc(reporter_class))

import dexy.template
DEFAULT_TEMPLATE = 'dexy:default'
def gen_command(
        d=None,  # The directory to place generated files in, must not exist.
        t=False, # Shorter alternative to --template.
        template=DEFAULT_TEMPLATE, # The alias of the template to use.
        **kwargs # Additional kwargs passed to template's run() method.
        ):
    """
    Generate a new dexy project in the specified directory, using the template.
    """
    if t and (template == DEFAULT_TEMPLATE):
        template = t
    elif t and template != DEFAULT_TEMPLATE:
        raise dexy.exceptions.UserFeedback("Only specify one of --t or --template, not both.")

    if not template in dexy.template.Template.aliases:
        print "Can't find a template named '%s'. Run 'dexy templates' for a list of templates." % template
        sys.exit(1)

    template_class = dexy.template.Template.aliases[template]
    template_class().run(d, **kwargs)

    # We run dexy setup. This will respect any dexy.conf file in the template
    # but passing command line options for 'setup' to 'gen' currently not supported.
    os.chdir(d)
    wrapper = init_wrapper({})
    wrapper.setup_dexy_dirs()
    print "Success! Your new dexy project has been created in directory '%s'" % d
    if os.path.exists("README"):
        with open("README", "r") as f:
            print f.read()
        print "\nThis information is in the 'README' file for future reference."

def template_command(
        alias=None
        ):
    print template_text(alias)

def template_text(
        alias=None
    ):
    template = dexy.template.Template.aliases[alias]
    for batch in template.dexy():
        man_doc_key = 'Doc:dexy.rst|jinja|rst2man'
        if man_doc_key in batch.lookup_table:
            man_doc = batch.lookup_table[man_doc_key].output().storage.data_file()

            command = "man %s" % man_doc
            import subprocess
            proc = subprocess.Popen(
                       command,
                       shell=True,
                       stdout=subprocess.PIPE,
                       stderr=subprocess.STDOUT
                   )
            stdout, stderr = proc.communicate()
            return stdout
        else:
            return "no example found"

def templates_command(
        simple=False, # Only print template names, without docstring or headers.
        validate=False # For developer use only, validate templates (runs and checks each template).
        ):
    """
    List templates that can be used to generate new projects.
    """
    aliases = []
    for alias in sorted(dexy.template.Template.aliases):
        klass = dexy.template.Template.aliases[alias]
        if klass.is_active():
            aliases.append(alias)

    if simple:
        print "\n".join(aliases)
    else:
        FMT = "%-40s %s"
        print FMT % ("Alias", "Info")
        for alias in aliases:
            klass = dexy.template.Template.aliases[alias]
            print FMT % (alias, getdoc(klass)),
            if validate:
                print " validating...",
                print klass.validate() and "OK" or "ERROR"
            else:
                print ''

        if len(aliases) == 1:
            print "Run '[sudo] pip install dexy-templates' to install some more templates."

        print "Run 'dexy help -on gen' for help on generating projects from templates."

import SimpleHTTPServer
import SocketServer
from dexy.plugins.website_reporters import WebsiteReporter
from dexy.plugins.output_reporters import OutputReporter
NO_OUTPUT_MSG = """Please run dexy first, or specify a directory to serve. \
For help run 'dexy help -on serve'"""

def serve_command(
        port=8085,
        directory=False
        ):
    """
    Runs a simple web server on dexy-generated files. Will look first to see if
    the Website Reporter has run, if so this content is served. If not the
    standard output/ directory contents are served. You can also specify
    another directory to be served. The port defaults to 8085, this can also be
    customized.

    """
    if not directory:
        if os.path.exists(WebsiteReporter.REPORTS_DIR):
            directory = WebsiteReporter.REPORTS_DIR
        elif os.path.exists(OutputReporter.REPORTS_DIR):
            directory = OutputReporter.REPORTS_DIR
        else:
            print NO_OUTPUT_MSG
            sys.exit(1)

    os.chdir(directory)

    Handler = SimpleHTTPServer.SimpleHTTPRequestHandler
    httpd = SocketServer.TCPServer(("", port), Handler)
    print "serving contents of %s on http://localhost:%s" % (directory, port)
    print "type ctrl+c to stop"
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        sys.exit(1)

INFO_ATTRS = [
        'name',
        'ext',
        'key',
        'hashstring',
        'storage_type'
        ]
INFO_METHODS = [
        'basename',
        'filesize',
        'baserootname',
        'parent_dir',
        'long_name',
        'web_safe_document_key'
        ]
STORAGE_METHODS = [
        'data_file',
        'data_file_exists'
        ]
def info_command(
        __cli_options=False,
        k=None, # The doc key to query. Use dexy grep to search doc keys.
        artifactsdir=D['artifacts_dir'], # location of directory in which to store artifacts
        logdir=D['log_dir'] # location of directory in which to store logs
        ):
    wrapper = init_wrapper(locals())
    wrapper.setup_read()
    data = wrapper.db.find_data_by_doc_key(k)

    print k

    print "  attributes:"
    for fname in sorted(INFO_ATTRS):
        print "    %s: %s" % (fname, getattr(data, fname))

    print "  methods:"
    for fname in sorted(INFO_METHODS):
        print "    %s: %s" % (fname, getattr(data, fname)())

    print "  storage methods:"
    for fname in sorted(STORAGE_METHODS):
        print "    %s: %s" % (fname, getattr(data.storage, fname)())



from dexy.doc import Doc
from dexy.tests.utils import wrap
from dexy.wrapper import Wrapper
import dexy.data
import dexy.exceptions
import os

def test_canonical_name():
    with wrap() as wrapper:
        doc = Doc("hello.txt",
                contents="hello",
                canonical_name="yello.abc",
                wrapper=wrapper)

        wrapper.run_docs(doc)
        assert doc.output().name == "yello.abc"
        wrapper.report()
        assert os.path.exists(os.path.join('output', 'yello.abc'))

def test_attempt_write_outside_project_root():
    with wrap() as wrapper:
        doc = Doc("../../example.txt",
                contents = "hello",
                wrapper=wrapper)

        try:
            wrapper.run_docs(doc)
            wrapper.report()
            assert False, 'should raise UserFeedback'
        except dexy.exceptions.UserFeedback as e:
            assert 'trying to write' in str(e)

def test_key_value_data():
    with wrap() as wrapper:
        data = dexy.data.KeyValue("doc.json", ".json", "doc.json", "hash1", {}, wrapper, storage_type='json')

        assert not data._data
        assert data.storage._data == {}

        # We use the append method to add key-value pairs.
        data.append('foo', 'bar')
        assert len(data.keys()) == 1

        assert data.value('foo') == 'bar'
        assert data.storage['foo'] == 'bar'
        assert data.as_text() == "foo: bar"
        data.as_sectioned()['foo'] == 'bar'

def test_key_value_data_sqlite():
    with wrap() as wrapper:
        data = dexy.data.KeyValue("doc.sqlite3", ".sqlite3", "doc.sqlite3", "hash1", {}, wrapper)

        data.append('foo', 'bar')
        assert len(data.keys()) == 1

        assert data.value('foo') == 'bar'
        assert ["%s: %s" % (k, v) for k, v in data.storage][0] == "foo: bar"

        data.as_text() == "foo: bar"

def test_generic_data():
    with wrap() as wrapper:
        CONTENTS = "contents go here"

        # Create a GenericData object
        data = dexy.data.Generic("doc.txt", ".txt", "doc.txt", "hash1", {}, wrapper)

        # Assign some text contents
        data._data = CONTENTS
        assert data.has_data()
        assert not data.is_cached()
        assert not data.filesize()

        # Save data to disk
        data.save()
        assert data.has_data()
        assert data.is_cached()
        assert data.filesize() > 10

        # Clear data from memory
        data._data = None
        assert data.has_data()

        # Load it again from disk
        data.load_data()
        assert data._data == CONTENTS

        # The convenience methods load from disk if needed.
        data._data = None
        assert data.as_text() == CONTENTS

        data._data = None
        assert data.as_sectioned()['1'] == CONTENTS

def test_init_data():
    wrapper = Wrapper()
    data = dexy.data.Generic("doc.txt", ".abc", "doc.abc", "def123", {}, wrapper)

    assert data.key == "doc.txt"
    assert data.name == "doc.abc"
    assert data.ext == ".abc"
    assert data.hashstring == "def123"

    assert not data.has_data()
    assert not data.is_cached()



from dexy.doc import Doc
from dexy.tests.utils import tempdir
from dexy.tests.utils import wrap
from dexy.wrapper import Wrapper

def test_yamlargs_with_caching():
    with tempdir():
        wrapper = Wrapper()
        wrapper.setup_dexy_dirs()
        doc = Doc("example.txt|yamlargs",
                contents = "title: My Title\n---\r\nThis is the content.",
                wrapper=wrapper)
        wrapper.run_docs(doc)

        task = wrapper.batch.lookup_table["FilterArtifact:example.txt|yamlargs"]
        assert task.args['title'] == "My Title"
        assert task.content_source == 'generated'

        wrapper = Wrapper()
        doc = Doc("example.txt|yamlargs",
                contents = "title: My Title\n---\r\nThis is the content.",
                wrapper=wrapper)
        wrapper.run_docs(doc)
        task = wrapper.batch.lookup_table["FilterArtifact:example.txt|yamlargs"]
        assert task.args['title'] == "My Title"
        assert task.content_source == 'cached'

def test_yamlargs_no_yaml():
    with wrap() as wrapper:
        doc = Doc("example.txt|yamlargs",
                contents = "This is the content.",
                wrapper=wrapper)

        wrapper.run_docs(doc)
        assert doc.output().as_text() == "This is the content."

def test_yamlargs():
    with wrap() as wrapper:
        doc = Doc("example.txt|yamlargs",
                contents = "title: My Title\n---\r\nThis is the content.",
                wrapper=wrapper)

        wrapper.run_docs(doc)
        assert doc.title() == "My Title"
        assert doc.output().as_text() == "This is the content."

YAML = """filterargs:
  foo: 5
  bar: 42
  baz: xyz
"""

RESULT = """Here are the arguments you passed:
bar: 42
baz: xyz
foo: 5"""
def test_yamlargs_filterargs():
    with wrap() as wrapper:
        doc = Doc("example.txt|yamlargs|filterargs",
                contents = "%s\n---\r\nThis is the content." % YAML,
                wrapper=wrapper)

        wrapper.run_docs(doc)
        assert doc.output().as_text() == RESULT



from dexy.tests.utils import runfilter
import dexy.exceptions

def test_tikz():
    with runfilter('tikz', TIKZ) as doc:
        assert ".pdf" in doc.output().name
        assert doc.output().is_cached()

def test_latex():
    with runfilter('latex', LATEX) as doc:
        assert ".pdf" in doc.output().name
        assert doc.output().is_cached()

def test_xetex():
    with runfilter('xetex', LATEX) as doc:
        assert ".pdf" in doc.output().name
        assert doc.output().is_cached()

def test_broken_latex():
    try:
        with runfilter('latex', BROKEN_LATEX):
            pass
        assert False, 'should raise UserFeedback'
    except dexy.exceptions.UserFeedback as e:
        assert "Latex file not generated" in e.message

TIKZ = """\
\\tikz \draw (0,0) -- (1,1)
{[rounded corners] -- (2,0) -- (3,1)}
-- (3,0) -- (2,1);
"""
LATEX = """\
\documentclass{article}
\\title{Hello, World!}
\\begin{document}
\maketitle
Hello!
\end{document}
"""

BROKEN_LATEX = """\
\documentclass{article}
"""



import time

class Batch(object):
    def __init__(self, wrapper, batch_id=None):
        self.wrapper = wrapper

        if batch_id:
            self.load(batch_id)
        else:
            self.new()

    def filters_used(self):
        filters = set()
        for doc in self.docs():
            filters = filters.union([f for f in doc.filters if not f.startswith("-")])
        return filters

    def load(self, batch_id):
        """
        Loads a previously run batch into memory.
        """
        self.batch_id = batch_id

    def save(self):
        """
        Persist a batch's info to disk (for later retrieval by load).
        """
        pass

    def add_doc(self, new_doc):
        self.lookup_table[new_doc.key_with_class()] = new_doc

    def new(self):
        """
        Sets up defaults for a new batch.
        """
        self.state = 'new'
        self.batch_id = self.wrapper.db.next_batch_id()
        self.previous_batch_id = self.wrapper.db.calculate_previous_batch_id(self.batch_id)

    def create_lookup_table(self):
        """
        Creates a lookup table assuming the tree has already been populated.
        """
        self.lookup_table = dict((t.key_with_class(), t) for t in self.tree)

    def load_ast(self, ast):
        self.ast = ast
        self.tree, self.nodes = self.ast.walk()

        self.lookup_table = {}
        for task in self.nodes.values():
            self.lookup_table[task.key_with_class()] = task

    def nodes_for_target(self, target=None):
        if target:
            # Look for identical target in root-level nodes
            nodes = [n for n in self.tree if n.key == target]
            
            if not nodes:
                # Look for similar target in root-level nodes
                nodes = [n for n in self.tree if n.key.startswith(target)]

            if not nodes:
                # Look for identical target anywhere in tree
                nodes = [n for n in self.nodes.values() if n.key == target]
                # TODO sort nodes..

            if not nodes:
                # Look for similar target anywhere in tree
                nodes = [n for n in self.nodes.values() if n.key.startswith(target)]
                # TODO sort nodes..

        else:
            nodes = self.tree

        return nodes

    def run(self, target=None):
        self.start_time = time.time()

        nodes = self.nodes_for_target(target)

        self.state = 'populating'

        for node in nodes:
            for task in node:
                task()

        self.state = 'settingup'

        for node in nodes:
            for task in node:
                task()

        self.state = 'running'

        for node in nodes:
            for task in node:
                task()

        self.state = 'complete'

        self.end_time = time.time()

    def elapsed(self):
        return self.end_time - self.start_time

    # Methods for accessing lists of tasks
    def task(self, task_key):
        if not ":" in task_key:
            task_key = "Doc:%s" % task_key
        return self.lookup_table[task_key]

    def tasks(self):
        return self.lookup_table.values()

    def docs(self):
        return [v for k, v in self.lookup_table.iteritems() if k.startswith("Doc:")]

    def doc_names(self):
        return [doc.name for doc in self.docs()]

    def tasks_by_elapsed(self, n=10):
        return sorted(self.lookup_table.values(), key=lambda task: hasattr(task, 'doc') and task.elapsed or None, reverse=True)[0:n]



from dexy.filter import Filter
from dexy.tests.utils import runfilter
from nose.exc import SkipTest
from nose.tools import raises
from dexy.utils import s

def test_s():
    text = """This is some text
    which goes onto
    many lines and has
    indents at the start."""
    assert s(text) == 'This is some text which goes onto many lines and has indents at the start.'

class InactiveDexyFilter(Filter):
    """
    filter which is always inactive, for testing
    """
    NODOC = True
    ALIASES = ['inactive']
    @classmethod
    def is_active(klass):
        return False

@raises(SkipTest)
def test_inactive_filters_skip():
    with runfilter("inactive", "hello"):
        pass



from dexy.common import OrderedDict
from dexy.plugin import PluginMeta
import chardet
import dexy.storage
import dexy.utils
import dexy.wrapper
import os
import posixpath
import shutil

class Data:
    __metaclass__ = PluginMeta

    ALIASES = []
    DEFAULT_STORAGE_TYPE = 'generic'

    @classmethod
    def is_active(klass):
        return True

    @classmethod
    def retrieve(klass, data_type, key, ext, name, hashstring, args, wrapper, storage_type):
        """
        Method to retrieve a Data object based on info stored in database.

        Optional kwags are passed to a RunParams instance.
        """
        data_class = klass.aliases[data_type]
        return data_class(key, ext, name, hashstring, args, wrapper, storage_type)

    @classmethod
    def storage_class_alias(klass, file_ext):
        return klass.DEFAULT_STORAGE_TYPE

    def __repr__(self):
        return "Data('%s')" % (self.key)

    def __unicode__(self):
        if isinstance(self.data(), unicode):
            return self.data()
        elif not self.data():
            return unicode(None)
        else:
            if self.wrapper.encoding == 'chardet':
                encoding = chardet.detect(self.data())['encoding']
                if not encoding:
                    return self.data().decode("utf-8")
                else:
                    return self.data().decode(encoding)
            else:
                return self.data().decode(self.wrapper.encoding)

    def __str__(self):
        return str(unicode(self))

    def __init__(self, key, ext, canonical_name, hashstring, args, wrapper, storage_type=None):
        self.key = key
        self.ext = ext
        self.name = canonical_name
        self.hashstring = hashstring
        self.args = args
        self.wrapper = wrapper

        self.setup_storage(storage_type)

        self._data = None

        # allow doing custom setup in subclasses
        self.setup()

    def keys(self):
        return []

    def setup(self):
        pass

    def setup_storage(self, storage_type):
        self.storage_type = storage_type or self.storage_class_alias(self.ext)
        storage_class = dexy.storage.Storage.aliases[self.storage_type]
        self.storage = storage_class(self.hashstring, self.ext, self.wrapper)
        self.storage.check_location_is_in_project_dir(self.name)

    def parent_dir(self):
        return posixpath.dirname(self.name)

    def long_name(self):
        if "|" in self.key:
            return "%s%s" % (self.key.replace("|", "-"), self.ext)
        else:
            return self.name

    def basename(self):
        return posixpath.basename(self.name)

    def baserootname(self):
        """
        Returns basename stripped of file extension.
        """
        return posixpath.splitext(self.basename())[0]

    def web_safe_document_key(self):
        return self.long_name().replace("/", "--")

    def relative_path_to(self, relative_to):
        return posixpath.relpath(relative_to, self.parent_dir())

    def relative_refs(self, relative_to_file):
        doc_dir = posixpath.dirname(relative_to_file)
        return [
                posixpath.relpath(self.key, doc_dir),
                posixpath.relpath(self.long_name(), doc_dir),
                "/%s" % self.key,
                "/%s" % self.long_name()
        ]

    # Define functions that might get called on expectation of a string...

    def strip(self):
        return unicode(self).strip()
    
    def splitlines(self, arg=None):
        return unicode(self).splitlines(arg)

class Generic(Data):
    """
    Data type representing generic binary or text-based data.
    """
    ALIASES = ['generic']
    DEFAULT_STORAGE_TYPE = 'generic'

    def save(self):
        if isinstance(self._data, unicode):
            self.storage.write_data(self._data.encode("utf-8"))
        else:
            self.storage.write_data(self._data)

    def set_data(self, data):
        """
        Set data to the passed argument and persist data to disk.
        """
        self._data = data
        self.save()

    def load_data(self):
        self._data = self.storage.read_data()

    def has_data(self):
        return self._data or self.storage.data_file_exists()

    def is_cached(self):
        if self.storage.data_file_size() == 0:
            self.wrapper.log.debug("Data file for %s (%s) has size 0" % (self.key, self.storage.data_file()))
        return self.storage.data_file_exists()

    def filesize(self):
        if self.is_cached():
            return os.path.getsize(self.storage.data_file())

    def data(self):
        if not self._data:
            self.load_data()
        return self._data

    def as_text(self):
        return unicode(self)

    def as_sectioned(self):
        return {'1' : self.data()}

    def json_as_dict(self):
        return self.from_json()

    def from_json(self):
        """
        Attempts to load data using a JSON parser, returning whatever objects are defined in the JSON.
        """
        if self._data:
            return dexy.utils.parse_json(self._data)
        else:
            with open(self.storage.data_file(), "r") as f:
                return dexy.utils.parse_json_from_file(f)

    def copy_from_file(self, filename):
        shutil.copyfile(filename, self.storage.data_file())

    def clear_data(self):
        self._data = None

    def clear_cache(self):
        if os.path.exists(self.storage.data_file()):
            os.remove(self.storage.data_file())

    def output_to_file(self, filepath):
        """
        Write canonical output to a file. Parent directory must exist already.
        """
        if not self.storage.copy_file(filepath):
            self.storage.write_data(self.data(), filepath)

class Sectioned(Generic):
    """
    Data in sections which must be kept in order.
    """
    ALIASES = ['sectioned']
    DEFAULT_STORAGE_TYPE = 'jsonordered'


    def __unicode__(self):
        return u"\n".join(unicode(v) for v in self.data().values())

    def __str__(self):
        return "\n".join(str(v) for v in self.data().values())

    def as_sectioned(self):
        return self.data()

    def keys(self):
        return self.data().keys()

    def output_to_file(self, filepath):
        """
        Write canonical output to a file.
        """
        with open(filepath, "wb") as f:
            f.write(unicode(self).encode("utf-8"))

    def value(self, key):
        return self.data()[key]

    def __getitem__(self, key):
        return self.value(key)

class KeyValue(Generic):
    """
    Data class for key-value data.
    """
    ALIASES  = ['keyvalue']
    DEFAULT_STORAGE_TYPE = 'sqlite3'

    @classmethod
    def storage_class_alias(klass, file_ext):
        if file_ext == '.sqlite3':
            return 'sqlite3'
        elif file_ext == '.json':
            return 'json'
        else:
            return klass.DEFAULT_STORAGE_TYPE

    def setup(self):
        self.storage.setup()

    def as_text(self):
        text = []
        for k, v in self.storage:
            text.append(u"%s: %s" % (k, v))
        return u"\n".join(text)

    def as_sectioned(self):
        od = OrderedDict()
        for k, v in self.storage:
            od[k] = v
        return od

    def data(self):
        return self.as_sectioned()

    def value(self, key):
        return self.storage[key]

    def __getitem__(self, key):
        return self.value(key)

    def append(self, key, value):
        self.storage.append(key, value)

    def query(self, query):
        return self.storage.query(query)

    def keys(self):
        return self.storage.keys()

    def save(self):
        self.storage.save()



from dexy.common import OrderedDict
from dexy.plugin import PluginMeta
import dexy.exceptions
import os
import shutil
import sqlite3

# Generic Data

class Storage:
    __metaclass__ = PluginMeta

    @classmethod
    def is_active(klass):
        return True

    def check_location_is_in_project_dir(self, filepath):
        project_root = os.path.abspath(os.getcwd())
        if not project_root in os.path.abspath(filepath):
            raise dexy.exceptions.UserFeedback("trying to write '%s' outside of '%s'" % (filepath, project_root))

    def __init__(self, hashstring, ext, wrapper):
        self.hashstring = hashstring
        self.ext = ext
        self.wrapper = wrapper

class GenericStorage(Storage):
    """
    Default type of storage where content is stored in files.
    """
    ALIASES = ['generic']

    def data_file(self):
        ext = self.ext
        if not ext:
            ext = ".ext"

        return os.path.join(self.wrapper.artifacts_dir, "%s%s" % (self.hashstring, ext))

    def data_file_exists(self):
        return os.path.exists(self.data_file())

    def data_file_size(self):
        if self.data_file_exists():
            return os.path.getsize(self.data_file())

    def write_data(self, data, filepath=None):
        if not filepath:
            filepath = self.data_file()

        self.check_location_is_in_project_dir(filepath)

        if self.data_file_exists() and not filepath == self.data_file():
            shutil.copyfile(self.data_file(), filepath)
        else:
            with open(filepath, "wb") as f:
                f.write(data)

    def read_data(self):
        with open(self.data_file(), "rb") as f:
            return f.read()

    def copy_file(self, filepath):
        """
        If data file exists, copy file and return true. Otherwise return false.
        """
        if self.data_file_exists():
            self.check_location_is_in_project_dir(filepath)
            shutil.copyfile(self.data_file(), filepath)
            return True
        else:
            return False

# Sectioned Data
import json
class JsonOrderedStorage(GenericStorage):
    """
    Storage for ordered sectional data using JSON.
    """
    ALIASES = ['jsonordered']
    MAX_DATA_DICT_DECIMALS = 5
    MAX_DATA_DICT_LENGTH = 10 ** MAX_DATA_DICT_DECIMALS

    @classmethod
    def convert_numbered_dict_to_ordered_dict(klass, numbered_dict):
        ordered_dict = OrderedDict()
        for x in sorted(numbered_dict.keys()):
            k = x.split(":", 1)[1]
            ordered_dict[k] = numbered_dict[x]
        return ordered_dict

    @classmethod
    def convert_ordered_dict_to_numbered_dict(klass, ordered_dict):
        if len(ordered_dict) >= klass.MAX_DATA_DICT_LENGTH:
            exception_msg = """Your data dict has %s items, which is greater than the arbitrary limit of %s items.
You can increase this limit by changing MAX_DATA_DICT_DECIMALS."""
            raise Exception(exception_msg % (len(ordered_dict), klass.MAX_DATA_DICT_LENGTH))

        data_dict = {}
        i = -1
        for k, v in ordered_dict.iteritems():
            i += 1
            fmt = "%%0%sd:%%s" % klass.MAX_DATA_DICT_DECIMALS
            data_dict[fmt % (i, k)] = v
        return data_dict

    def value(self, key):
        return self.data()[key]

    def __getitem__(self, key):
        return self.value(key)

    def read_data(self):
        with open(self.data_file(), "rb") as f:
            numbered_dict = json.load(f)
            return self.convert_numbered_dict_to_ordered_dict(numbered_dict)

    def write_data(self, data, filepath=None):
        if not filepath:
            filepath = self.data_file()

        self.check_location_is_in_project_dir(filepath)

        with open(filepath, "wb") as f:
            numbered_dict = self.convert_ordered_dict_to_numbered_dict(data)
            json.dump(numbered_dict, f)

# Key Value Data
class JsonStorage(GenericStorage):
    """
    Storage for key value data using JSON.
    """
    ALIASES = ['json']

    def setup(self):
        self._data = {}

    def append(self, key, value):
        self._data[key] = value

    def keys(self):
        return self.data().keys()

    def value(self, key):
        return self.data()[key]

    def __getitem__(self, key):
        return self.value(key)

    def __iter__(self):
        for k, v in self.data().iteritems():
            yield k, v

    def read_data(self):
        with open(self.data_file(), "rb") as f:
            return json.load(f)

    def data(self):
        if len(self._data) == 0:
            self._data = self.read_data()
        return self._data

    def write_data(self, data, filepath=None):
        if not filepath:
            filepath = self.data_file()

        self.check_location_is_in_project_dir(filepath)

        with open(filepath, "wb") as f:
            json.dump(data, f)

    def save(self):
        with open(self.data_file(), "wb") as f:
            json.dump(self._data, f)

class Sqlite3Storage(GenericStorage):
    """
    Storage of key value storage in sqlite3 database files.
    """
    ALIASES = ['sqlite3']

    def working_file(self):
        return os.path.join(self.wrapper.artifacts_dir, "%s-tmp%s" % (self.hashstring, self.ext))

    def setup(self):
        if os.path.exists(self.data_file()):
            self._storage = sqlite3.connect(self.data_file())
            self._cursor = self._storage.cursor()
        else:
            if os.path.exists(self.working_file()):
                os.remove(self.working_file())
            self._storage = sqlite3.connect(self.working_file())
            self._cursor = self._storage.cursor()
            self._cursor.execute("CREATE TABLE kvstore (key TEXT, value TEXT)")
            self._append_counter = 0

    def append(self, key, value):
        self._cursor.execute("INSERT INTO kvstore VALUES (?, ?)", (str(key), str(value)))
        self._append_counter += 1
        if self._append_counter > 1000:
            self.wrapper.log.debug("intermediate commit to sqlite db, resetting append counter to 0")
            self._storage.commit()
            self._append_counter = 0

    def keys(self):
        self._cursor.execute("SELECT key from kvstore")
        return [str(k[0]) for k in self._cursor.fetchall()]

    def value(self, key):
        self._cursor.execute("SELECT value from kvstore where key = ?", (key,))
        return self._cursor.fetchone()[0]

    def query(self, query):
        if not '%' in query:
            query = "%%%s%%" % query
        self._cursor.execute("SELECT * from kvstore where key like ?", (query,))
        return self._cursor.fetchall()

    def __getitem__(self, key):
        return self.value(key)

    def __iter__(self):
        self._cursor = self._storage.cursor()
        self._cursor.execute("SELECT * from kvstore")
        rows = self._cursor.fetchall()
        for k, v in rows:
            yield k, v

    def save(self):
        self.check_location_is_in_project_dir(self.data_file())
        self._storage.commit()
        shutil.copyfile(self.working_file(), self.data_file())



from dexy.filter import DexyFilter
from dexy.plugins.process_filters import SubprocessFilter

try:
    from docutils import core
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class RestructuredTextFilter(DexyFilter):
    """
    A 'native' ReST filter which uses the docutils library.
    """
    ALIASES = ['rst']
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".html", ".tex", ".xml"]

    @classmethod
    def is_active(klass):
        return AVAILABLE

    def process_text(self, input_text):
        if self.artifact.ext == ".html":
            parts = core.publish_parts(
                    input_text,
                    writer_name = "html"
                    )
            return parts['body']
        elif self.artifact.ext == ".tex":
            parts = core.publish_parts(
                    input_text,
                    writer_name = "latex"
                    )

            # Note any latex requirements in logfile
            self.log.debug("Requirements for ReST:")
            for l in parts['requirements'].splitlines():
                self.log.debug(l)
            return parts['body']
        else:
            raise Exception("unsupported extension %s" % self.artifact.ext)

class Rst2HtmlFilter(SubprocessFilter):
    """
    This uses the command line tool rst2html.
    """
    ADD_NEW_FILES = False
    EXECUTABLES = ['rst2html', 'rst2html.py']
    VERSION_COMMAND = 'rst2html.py --version'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".html"]
    ALIASES = ['rst2html']

class Rst2LatexFilter(SubprocessFilter):
    """
    This uses the command line tool rst2latex.
    """
    ADD_NEW_FILES = False
    ALIASES = ['rst2latex']
    EXECUTABLES = ['rst2latex', 'rst2latex.py']
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".tex"]
    VERSION_COMMAND = 'rst2latex.py --version'

class Rst2XmlFilter(SubprocessFilter):
    """
    This uses the command line tool rst2xml.
    """
    ADD_NEW_FILES = False
    EXECUTABLES = ['rst2xml', 'rst2xml.py']
    VERSION_COMMAND = 'rst2xml.py --version'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".tex"]
    ALIASES = ['rst2xml']

class Rst2OdtFilter(SubprocessFilter):
    """
    This uses the command line tool rst2odt.
    """
    EXECUTABLES = ['rst2odt', 'rst2odt.py']
    VERSION_COMMAND = 'rst2odt.py --version'
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".odt"]
    ALIASES = ['rst2odt']

class Rst2BeamerFilter(SubprocessFilter):
    """
    Filter for rst2beamer command line tool, requires docutils plus rst2beamer package.
    """
    ADD_NEW_FILES = False
    ALIASES = ['rst2beamer']
    EXECUTABLES = ['rst2beamer', 'rst2beamer.py']
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".tex"]
    VERSION_COMMAND = "rst2beamer --version"

class Rst2Man(SubprocessFilter):
    """
    Filter for rst2man command line tool, requires docutils.
    """
    ADD_NEW_FILES = False
    ALIASES = ['rst2man']
    EXECUTABLES = ['rst2man', 'rst2man.py']
    INPUT_EXTENSIONS = [".rst", ".txt"]
    OUTPUT_EXTENSIONS = [".man"]
    VERSION_COMMAND = "rst2man.py --version"



import dexy.exceptions
import inspect
import json
import os
import yaml
import posixpath
import re

def value_for_hyphenated_or_underscored_arg(arg_dict, arg_name_hyphen, default=None):
    if not "-" in arg_name_hyphen and "_" in arg_name_hyphen:
        raise dexy.exceptions.InternalDexyProblem("arg_name_hyphen %s has underscores!" % arg_name_hyphen)

    arg_name_underscore = arg_name_hyphen.replace("-", "_")

    arg_value = arg_dict.get(arg_name_hyphen)

    if arg_value is None:
        arg_value = arg_dict.get(arg_name_underscore)

    if arg_value is None:
        return default
    else:
        return arg_value

def s(text):
    return re.sub('\s+', ' ', text)

def getdoc(element, firstline=True):
    docstring = inspect.getdoc(element)
    if docstring and firstline:
        docstring = docstring.splitlines()[0]

    if not docstring:
        docstring = ''

    return docstring

def os_to_posix(path):
    return posixpath.join(*os.path.split(path))

def parse_json(input_text):
    try:
        return json.loads(input_text)
    except ValueError as e:
        msg = inspect.cleandoc("""Was unable to parse the JSON you supplied.
        Here is information from the JSON parser:""")
        msg += "\n"
        msg += str(e)
        raise dexy.exceptions.UserFeedback(msg)

def parse_json_from_file(f):
    try:
        return json.load(f)
    except ValueError as e:
        msg = inspect.cleandoc("""Was unable to parse the JSON you supplied.
        Here is information from the JSON parser:""")
        msg += "\n"
        msg += str(e)
        raise dexy.exceptions.UserFeedback(msg)

def parse_yaml(input_text):
    """
    Parse a single YAML document.
    """
    try:
        return yaml.load(input_text)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as e:
        msg = inspect.cleandoc("""Was unable to parse the YAML you supplied.
        Here is information from the YAML parser:""")
        msg += "\n"
        msg += str(e)
        raise dexy.exceptions.UserFeedback(msg)

def parse_yamls(input_text):
    """
    Parse YAML content that may include more than 1 document.
    """
    try:
        return yaml.load_all(input_text)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as e:
        msg = inspect.cleandoc("""Was unable to parse the YAML you supplied.
        Here is information from the YAML parser:""")
        msg += "\n"
        msg += str(e)
        raise dexy.exceptions.UserFeedback(msg)

def printable_for_char(c):
    if ord(c) >= ord('!'):
        return c
    elif ord(c) == 32:
        return "<space>"
    else:
        return ""

def char_diff(str1, str2):
    """
    Returns a char-by-char diff of two strings, highlighting differences.
    """
    msg = ""
    for i, c1 in enumerate(str1):
        if len(str2) > i:
            c2 = str2[i]

            if c1 == c2:
                flag = ""
            else:
                flag = " <---"

            p_c1 = printable_for_char(c1)
            p_c2 = printable_for_char(c2)

            msg = msg + "\n%5d: %8s\t%8s\t\t%s\t%s %s" % (i, p_c1, p_c2, ord(c1), ord(c2), flag)
        else:
            # str1 is longer than str2
            flag = " <---"
            p_c1 = printable_for_char(c1)
            msg = msg + "\n%5d: %8s\t%8s\t\t%s\t%s %s" % (i, p_c1, '  ', ord(c1), '  ', flag)

    # TODO add code for str2 longer than str1

    return msg

# Based on http://nedbatchelder.com/code/utilities/wh.py
def command_exists(cmd_name):
    path = os.environ["PATH"]
    if ";" in path:
        path = filter(None, path.split(";"))
    else:
        path = filter(None, path.split(":"))

    if os.environ.has_key("PATHEXT"):
        # Present on windows, returns e.g. '.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC'
        pathext = os.environ["PATHEXT"]
        pathext = filter(None, pathext.split(";"))
    else:
        # Not windows, look for exact command name.
        pathext = [""]

    cmd_found = False
    for d in path:
        for e in pathext:
            filepath = os.path.join(d, cmd_name + e)
            if os.path.exists(filepath):
                cmd_found = True
                break
    return cmd_found



from dexy.node import DocNode
from dexy.plugins.templating_filters import TemplateFilter
from dexy.plugins.templating_plugins import TemplatePlugin
from dexy.tests.utils import run_templating_plugin as run
from dexy.tests.utils import wrap
import dexy.plugins.templating_plugins as plugin
import inspect
import os

def test_base():
    run(TemplatePlugin)

def test_ppjson():
    with run(plugin.PrettyPrintJson) as env:
        assert 'ppjson' in env
        assert hasattr(env['ppjson'], '__call__')

def test_python_datetime():
    with run(plugin.PythonDatetime) as env:
        assert env['cal'].__class__.__name__ == 'Calendar'

def test_dexy_version():
    with run(plugin.DexyVersion) as env:
        assert env['DEXY_VERSION']

def test_simple_json():
    with run(plugin.SimpleJson) as env:
        assert inspect.ismodule((env['json']))

def test_python_builtins():
    with run(plugin.PythonBuiltins) as env:
        assert 'hasattr' in env

def test_pygments():
    with run(plugin.PygmentsStylesheet) as env:
        assert 'pastie.tex' in env['pygments'].keys()
        assert 'pastie.css' in env['pygments'].keys()
        assert 'pastie.html' in env['pygments'].keys()
        assert hasattr(env['highlight'], '__call__')

class TestSubdirectory(TemplateFilter):
    """
    test subdir
    """
    ALIASES = ['testsubdir']
    TEMPLATE_PLUGINS = [plugin.Subdirectories]

def test_subdirectories():
    with wrap() as wrapper:
        os.makedirs("s1")
        os.makedirs("s2")

        node = DocNode("file.txt|testsubdir",
                contents="hello",
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]

        env = doc.final_artifact.filter_instance.run_plugins()
        assert 's1' in env['subdirectories']
        assert 's2' in env['subdirectories']

class TestVariables(TemplateFilter):
    """
    test variables
    """
    ALIASES = ['testvars']
    TEMPLATE_PLUGINS = [plugin.Variables]

def test_variables():
    with wrap() as wrapper:
        node = DocNode("hello.txt|testvars",
                contents = "hello",
                testvars = { "variables" : {"foo" : "bar", "x" : 123.4 } },
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]

        env = doc.final_artifact.filter_instance.run_plugins()
        assert env['foo'] == 'bar'
        assert env['x'] == 123.4

class TestGlobals(TemplateFilter):
    """
    test globals
    """
    ALIASES = ['testglobals']
    TEMPLATE_PLUGINS = [plugin.Globals]

def test_globals():
    with wrap() as wrapper:
        wrapper.globals = "foo=bar"
        node = DocNode("hello.txt|testglobals",
                contents = "hello",
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        env = doc.final_artifact.filter_instance.run_plugins()
        assert env['foo'] == 'bar'



from dexy.doc import Doc
from dexy.tests.utils import wrap
from nose.tools import raises
import dexy.exceptions

@raises(dexy.exceptions.UserFeedback)
def test_idio_invalid_input():
    with wrap() as wrapper:
        doc = Doc("hello.py|idio", wrapper=wrapper, contents=" ")
        wrapper.run_docs(doc)

@raises(dexy.exceptions.UserFeedback)
def test_idio_bad_file_extension():
    with wrap() as wrapper:
        doc = Doc("hello.xyz|idio", wrapper=wrapper, contents=" ")
        wrapper.run_docs(doc)

def test_multiple_sections():
    with wrap() as wrapper:
        src = """
### @export "vars"
x = 6
y = 7

### @export "multiply"
x*y

"""
        doc = Doc("example.py|idio",
                contents=src,
                wrapper=wrapper)

        wrapper.run_docs(doc)

        assert doc.output().keys() == ['1', 'vars', 'multiply']

def test_force_text():
    with wrap() as wrapper:
        doc = Doc("example.py|idio|t",
                contents="print 'hello'\n",
                wrapper=wrapper)

        wrapper.run_docs(doc)

        assert str(doc.output()) == "print 'hello'\n"

def test_force_latex():
    with wrap() as wrapper:
        doc = Doc("example.py|idio|l",
                contents="print 'hello'\n",
                wrapper=wrapper)

        wrapper.run_docs(doc)

        assert "begin{Verbatim}" in str(doc.output())



from dexy.filter import DexyFilter
from dexy.utils import parse_yaml
import re

class YamlargsFilter(DexyFilter):
    """
    Strips YAML metadata from top of file and adds this to args for every artifact in current doc.
    """
    ALIASES = ['yamlargs']

    def process_text(self, input_text):
        regex = "\r?\n---\r?\n"
        if re.search(regex, input_text):
            yamlargs, content = re.split(regex, input_text)
            args = parse_yaml(yamlargs)
            for a in [self.artifact.doc] + self.artifact.doc.children:
                a.args.update(args)
            return content
        else:
            return input_text



from dexy.plugin import PluginMeta
from dexy.utils import os_to_posix
import StringIO
import dexy.doc
import dexy.exceptions
import logging

class Task():
    ALIASES = []
    __metaclass__ = PluginMeta

    STATE_TRANSITIONS = [
            ('new', 'populating'),
            ('populating', 'populated'),
            ('populated', 'settingup'),
            ('settingup', 'setup'),
            ('setup', 'running'),
            ('running', 'complete'),

            # sometimes want to skip directly to populated
            ('new', 'populated')
            ]

    @classmethod
    def create(klass, alias, pattern, **kwargs):
        task_class = klass.aliases[alias]
        return task_class(pattern, **kwargs)

    def key_for_log(self):
        return self.key

    def to_arg(self):
        alias = self.ALIASES[0]
        return "%s:%s" % (alias, self.key)

    def __repr__(self):
        return self.key_with_class()

    def __init__(self, key, **args):
        self.key = os_to_posix(key)
        self.args = args
        self.args_before_defaults = args
        self.state = 'new'
        self.elapsed = None
        self.children = []
        self.inputs = []
        self.created_by_doc = None
        self.metadata = dexy.metadata.Md5()

        if args.has_key('wrapper') and args['wrapper']:
            self.wrapper = args['wrapper']

    def set_hashstring(self):
        self.hashstring = self.metadata.compute_hash()

    def transition(self, to_state):
        if (self.state, to_state) in self.STATE_TRANSITIONS:
            self.state = to_state
        else:
            raise dexy.exceptions.InvalidStateTransition("%s => %s" % (self.state, to_state))

    def __iter__(self):
        def next_task():
            if self.state == 'new':
                self.transition('populating')
                yield self
                self.transition('populated')

            elif self.state == 'populated':
                self.transition('settingup')
                yield self
                self.transition('setup')

            elif self.state == 'setup':
                if self.wrapper.batch.state == 'running':
                    self.transition('running')
                    yield self.pre
                    yield self
                    yield self.post
                    self.transition('complete')

            elif self.state in ('running', 'populating', 'settingup',):
                raise dexy.exceptions.CircularDependency

            elif self.state == 'complete':
                pass

            else:
                raise dexy.exceptions.UnexpectedState("%s in %s" % (self.state, self.key))

        return next_task()

    def __call__(self, *args, **kw):
        for inpt in self.inputs:
            for task in inpt:
                task(*args, **kw)

        for child in self.children:
            for task in child:
                task(*args, **kw)

        if self.state == 'populating':
            self.populate()

        elif self.state == 'settingup':
            self.setup()
            self.wrapper.batch.add_doc(self)

        elif self.state == 'running':
            self.wrapper.db.add_task_before_running(self)
            self.run(*args, **kw)
            self.wrapper.db.update_task_after_running(self)

        else:
            raise dexy.exceptions.UnexpectedState("%s in %s" % (self.state, self.key))

    def setup(self):
        pass

    def run(self, *args, **kw):
        pass

    def populate(self):
        pass

    def pre(self, *args, **kw):
        pass

    def post(self, *args, **kw):
        pass

    def websafe_key(self):
        return self.key.replace("/", "--").replace("|", "--").replace("*","")

    def key_with_class(self):
        return "%s:%s" % (self.__class__.__name__, self.key)

    def key_with_batch_id(self):
        return "%s:%s" % (self.wrapper.batch.batch_id, self.key_with_class())

    def set_log(self):
        if not hasattr(self, 'log'):
            self.log = logging.getLogger(self.key_for_log())
            self.logstream = StringIO.StringIO()
            handler = logging.StreamHandler(self.logstream)
            if hasattr(self, 'wrapper'):
                handler.setFormatter(logging.Formatter(self.wrapper.log_format))
            self.log.addHandler(handler)
            self.log.setLevel(logging.DEBUG)

            try:
                self.log.addHandler(logging.getLogger('dexy').handlers[0])
            except IndexError:
                pass



from datetime import datetime
from dexy.plugins.output_reporters import OutputReporter
from dexy.plugins.templating_plugins import PythonBuiltins
from jinja2 import Environment
from jinja2 import FileSystemLoader
import dexy.exceptions
import jinja2
import os

class WebsiteReporter(OutputReporter):
    """
    Applies a template to create a website from your dexy output.

    Templates are applied to all files with .html extension which don't already
    contain "<head" or "<body" tags.

    Templates must be named _template.html with no dexy filters applied (TODO relax this)
    """
    ALIASES = ['ws']
    REPORTS_DIR = 'output-site'
    ALLREPORTS = False

    def nav_directories(self):
        """
        Returns a dict whose keys are top-level directores containing an
        'index.html' page and whose values are a list with the 'doc' object
        for the 'index.html' page and a dict of subdirectories in same format.

        """
        directories = [None, {}]

        self.log.debug("In nav_directories")

        def populate(keys, value):
            temp = directories
            for k in keys:
                if not temp[1]:
                    temp[1] = {}
                if not temp[1].has_key(k):
                    temp[1][k] = [None, {}]
                temp = temp[1][k]
            temp[0] = doc

        for doc in self.wrapper.batch.docs():
            doc_dir = doc.output().parent_dir()
            if doc.is_index_page():
                path_elements = os.path.split(doc_dir)

                while path_elements and path_elements[0] in ('', '.'):
                    path_elements = path_elements[1:]

                if not path_elements:
                    directories[0] = doc
                else:
                    self.log.debug("adding doc %s" % doc.key)
                    populate(path_elements, doc)
            else:
                self.log.debug("doc %s is not index page, skipping" % doc.key)

        return directories

    def apply_and_render_template(self, doc):
        ws_template = doc.args.get('ws_template')
        if ws_template and not isinstance(ws_template, bool):
            template_file = ws_template
        else:
            template_file = "_template.html"
        template_path = None

        path_elements = doc.output().parent_dir().split(os.sep)
        for i in range(len(path_elements), -1, -1):
            template_path = os.path.join(*(path_elements[0:i] + [template_file]))
            if os.path.exists(template_path):
                self.log.debug("  using template %s for %s" % (template_path, doc.key))
                break

        if not template_path:
            raise dexy.exceptions.UserFeedback("  no template path for %s" % doc.key)

        env = Environment(undefined=jinja2.StrictUndefined)
        env.loader = FileSystemLoader([".", os.path.dirname(template_path)])
        self.log.debug("  loading template at %s" % template_path)
        template = env.get_template(template_path)

        if doc.is_index_page():
            nav_current_index = doc.output().parent_dir()
        else:
            nav_current_index = None

        if doc.final_artifact.ext == '.html':
            content = unicode(doc.output())
        else:
            content = doc.output()

        navigation = {
                'current_index' : nav_current_index,
                'directories' : self._nav_directories
                }

        env_data = {
                'content' : content,
                'locals' : locals,
                'navigation' : navigation,
                'page_title' : doc.title(),
                'source' : doc.name,
                'template_source' : template_path,
                'wrapper' : self.wrapper,
                'year' : datetime.now().year
                }

        for builtin in PythonBuiltins.PYTHON_BUILTINS:
            env_data[builtin.__name__] = builtin

        env_data.update(self.wrapper.parse_globals())

        fp = os.path.join(self.REPORTS_DIR, doc.output().name).replace(".json", ".html")

        parent_dir = os.path.dirname(fp)
        if not os.path.exists(parent_dir):
            os.makedirs(os.path.dirname(fp))

        self.log.debug("  writing to %s" % (fp))
        template.stream(env_data).dump(fp, encoding="utf-8")

    def run(self, wrapper):
        self.wrapper=wrapper
        self.set_log()
        self.keys_to_outfiles = []
        self._nav_directories = self.nav_directories()

        self.create_reports_dir()

        for doc in wrapper.batch.docs():
            self.log.debug("processing doc %s" % doc.key_with_class())
            if doc.canon:
                if doc.final_artifact.ext == ".html":
                    fragments = ('<html', '<body', '<head')
                    has_html_header = any(html_fragment in unicode(doc.output()) for html_fragment in fragments)

                    if has_html_header and not doc.args.get('ws_template'):
                        self.log.debug("  found html tag in output of %s" % doc.key)
                        self.write_canonical_doc(doc)
                    else:
                        self.apply_and_render_template(doc)
                elif doc.final_artifact.ext == '.json' and 'htmlsections' in doc.filters:
                    self.apply_and_render_template(doc)
                else:
                    self.write_canonical_doc(doc)



from dexy.reporter import Reporter
import subprocess
import os

class GraphReporter(Reporter):
    """
    Reporter which prints a graph representation of tree.
    """
    ALIASES = ['graph', 'dot']
    ALLREPORTS = False

    @classmethod
    def write_graph_to_dotfile(klass, wrapper, dotfile):
        with open(dotfile, "w") as f:
            f.write(wrapper.batch.graph)

    @classmethod
    def render_dotfile_to_png(klass, dotfile, pngfile):
        command = "dot %s -Tpng -o%s" % (dotfile, pngfile)
        proc = subprocess.Popen(
                   command,
                   shell=True,
                   stdout=subprocess.PIPE,
                   stderr=subprocess.STDOUT
               )
        proc.communicate()

    def run(self, wrapper):
        dotfile = os.path.join(wrapper.log_dir, 'dexygraph.dot')
        pngfile = os.path.join(wrapper.log_dir, 'dexygraph.png')

        self.write_graph_to_dotfile(wrapper, dotfile)
        self.render_dotfile_to_png(dotfile, pngfile)



import dexy.filter
import json
import os

class ApiFilter(dexy.filter.DexyFilter):
    """
    Base class for filters which post content to a remote API, such as a blog, CMS
    or helpdesk app. Provides standard formats and locations for storing
    configuration and authentication information.

    Need to read config for the API in general, such as the base URL and an API
    key for authentication.

    Also need to read config for the particular task/document being uploaded.
    This could be stored in the .dexy config entry, but this has two drawbacks.
    First, it makes it difficult to bulk define documents according to a
    pattern. Secondly, it makes it very difficult for dexy to modify this
    configuration to add additional information, such as the returned id for a
    newly created document.

    So, it is preferable to define a local file (defined by a relative path to
    the document in question) which can be overridden per-document in which we
    just store the API-related config and which can be modified by dexy without
    concern about identifying the entry in a .dexy file or accidentally
    overwriting some unrelated information.
    """
    ALIASES = ['apis']
    MASTER_API_KEY_FILE = "~/.dexyapis"
    PROJECT_API_KEY_FILE = ".dexyapis"
    PAGE_CONTENT_EXTENSIONS = ['.md', '.txt', '.html']
    API_KEY_NAME = None
    API_KEY_KEYS = ['username', 'password', 'url']

    DOCUMENT_API_CONFIG_FILE = None
    DOCUMENT_API_CONFIG_FILE_KEY = "api-config-file"
    DOCUMENT_API_CONFIG_POSTFIX = "-config.json"

    # Put API key locations in this array, earlier entries have priority.
    API_KEY_LOCATIONS = [PROJECT_API_KEY_FILE, MASTER_API_KEY_FILE]

    @classmethod
    def docmd_create_keyfile(klass):
        """
        Creates a key file in location specified by MASTER_API_KEY_FILE.
        """
        key_filename = os.path.expanduser(klass.MASTER_API_KEY_FILE)
        if os.path.exists(key_filename):
            raise Exception("File %s already exists!" % key_filename)

        keyfile_content = {}
        for filter_class in dexy.filter.Filter.aliases.values():
            if issubclass(filter_class, klass) and not filter_class == klass:
                print filter_class.__name__
                if not filter_class.API_KEY_NAME:
                    raise Exception("filter class %s should set API_KEY_NAME" % filter_class.__name__)
                keyfile_content[filter_class.API_KEY_NAME] = dict((k, "TODO") for k in filter_class.API_KEY_KEYS)

        with open(key_filename, "wb") as f:
            json.dump(keyfile_content, f, sort_keys = True, indent=4)

    def document_api_config_file(self):
        return self.args().get(self.DOCUMENT_API_CONFIG_FILE_KEY, self.DOCUMENT_API_CONFIG_FILE)

    def document_config_file(self):
        postfix_config_filename = "%s%s" % (os.path.splitext(self.output().name)[0], self.DOCUMENT_API_CONFIG_POSTFIX)
        if os.path.exists(postfix_config_filename):
            return postfix_config_filename
        else:
            return os.path.join(self.output().parent_dir(), self.document_api_config_file())

    def read_document_config(self):
        document_config = self.document_config_file()
        if os.path.exists(document_config):
            with open(document_config, "r") as f:
                return json.load(f)
        else:
            msg = "Filter %s needs a file %s, couldn't find it."
            raise dexy.exceptions.UserFeedback(msg % (self.artifact.filter_alias, document_config))

    def save_document_config(self, config):
        document_config = self.document_config_file()
        with open(document_config, "w") as f:
            json.dump(config, f, sort_keys=True, indent=4)

    @classmethod
    def read_param_class(klass, param_name):
        param_value = None
        if not param_name in klass.API_KEY_KEYS:
            raise Exception("Param %s not specified in API_KEY_KEYS for %s" % (param_name, klass.__name__))

        for filename in klass.API_KEY_LOCATIONS:
            if "~" in filename:
                filename = os.path.expanduser(filename)

            if os.path.exists(filename):
                with open(filename, "r") as f:
                    params = json.load(f)
                    if params.has_key(klass.API_KEY_NAME):
                        param_value = params[klass.API_KEY_NAME].get(param_name)

            if param_value:
                break

        if param_value and isinstance(param_value, basestring) and param_value.startswith("$"):
            # need to get value of bash variable
            param_value_from_env = os.getenv(param_value.lstrip("$"))
            if not param_value_from_env:
                raise Exception("Bash variable %s not defined in this environment!" % param_value)
            param_value = param_value_from_env

        if param_value:
            return param_value
        else:
            msg = "Could not find %s for %s in: %s" % (param_name, klass.API_KEY_NAME, ", ".join(klass.API_KEY_LOCATIONS))
            raise Exception(msg)

    def read_param(self, param_name):
        param_value = self.arg_value(param_name)
        if not param_value:
            param_value = self.__class__.read_param_class(param_name)

        return param_value



from dexy.common import OrderedDict
from dexy.version import DEXY_VERSION
import dexy.data
import dexy.doc
import dexy.exceptions
import dexy.filter
import dexy.metadata
import dexy.node
import dexy.task
import hashlib
import inspect
import json
import os
import posixpath
import shutil
import stat
import time

class Artifact(dexy.task.Task):
    """
    Task classes representing steps in dexy processing.
    """
    def __init__(self, key, **kwargs):
        super(Artifact, self).__init__(key, **kwargs)
        self.created_by_doc = None
        self.remaining_doc_filters = []

    def key_for_log(self):
        if len(self.remaining_doc_filters) > 0:
            return "%s(|%s)" % (self.key, "|".join(self.remaining_doc_filters))
        else:
            return self.key

    def key_with_class(self):
        return "%s:%s" % (self.__class__.__name__, self.key_for_log())

    def tmp_dir(self):
        return os.path.join(self.wrapper.artifacts_dir, self.hashstring)

    def input_filename(self):
        if self.ext and (self.ext == self.prior.ext):
            return "%s-work%s" % (self.input_data.baserootname(), self.prior.ext)
        else:
            return self.input_data.basename()

    def output_filename(self):
        return self.output_data.basename()

    def canonical_name_from_args(self):
        raw_arg_name = self.doc.arg_value('canonical-name')

        if raw_arg_name:
            if "/" in raw_arg_name:
                return raw_arg_name
            else:
                return posixpath.join(posixpath.dirname(self.key), raw_arg_name)

    def calculate_canonical_name(self):
        from_args = self.canonical_name_from_args()
        if from_args:
            return from_args % self.args
        else:
            name_without_ext = posixpath.splitext(self.key)[0]
            return "%s%s" % (name_without_ext, self.ext)

    def working_dir(self):
        return os.path.join(self.tmp_dir(), self.output_data.parent_dir())

    def working_dir_exists(self):
        return os.path.exists(self.working_dir()) and os.path.isdir(self.working_dir())

    def setup_wd(self, input_filename):
        tmp_dir = self.tmp_dir()

        # Clear any old working dir.
        shutil.rmtree(tmp_dir, ignore_errors=True)

        # Create the base working dir.
        os.mkdir(tmp_dir)

        for doc in self.doc.node.walk_input_docs():
            if doc.state == 'complete' or len(doc.filters) == 0:
                import re
                exclude_wd = self.args.get('exclude_wd')
                if exclude_wd and re.search(exclude_wd, doc.key):
                    self.log.debug("not saving '%s' to wd for '%s' because matches exclude_wd pattern '%s'" % (doc.key, self.key, exclude_wd))
                else:
                    # Figure out path to write this child doc.
                    filename = os.path.join(tmp_dir, doc.output().name)
                    parent_dir = os.path.dirname(filename)

                    # Ensure all parent directories exist.
                    if not os.path.exists(parent_dir):
                        os.makedirs(parent_dir)

                    yield(doc, filename)

        parent_dir = self.working_dir()
        input_filepath = os.path.join(parent_dir, input_filename)

        if not os.path.exists(parent_dir):
            os.makedirs(parent_dir)

        self.input_data.output_to_file(input_filepath)

    def data_class_alias(self):
        return 'generic'

    def setup_output_data(self):
        data_class = dexy.data.Data.aliases[self.data_class_alias()]
        self.log.debug("setting up output data of class %s" % data_class.__name__)
        self.output_data_type = data_class.ALIASES[0]
        self.output_data = data_class(self.key, self.ext, self.calculate_canonical_name(), self.hashstring, self.args, self.wrapper)

    def setup(self):
        self.set_log()
        self.log.debug("setting up %s" % self.key_with_class())

        self.set_extension()

        self.metadata = dexy.metadata.Md5()
        self.set_metadata_attrs()

        # sets hashstring w/o node data, will be set again by node.
        self.set_hashstring()

        if hasattr(self, 'prior') and self.prior:
            self.input_data = self.prior.output_data
        self.setup_output_data()

class InitialArtifact(Artifact):
    """
    The artifact class representing an initial artifact, pretty much just a copy of the original file being processed.
    """
    def append_child_hashstrings(self):
        pass

    def set_metadata_attrs(self):
        self.metadata.key = self.key

        stat_info = os.stat(self.name)
        self.metadata.mtime = stat_info[stat.ST_MTIME]
        self.metadata.size = stat_info[stat.ST_SIZE]

    def set_output_data(self):
        self.output_data.copy_from_file(self.name)

    def set_extension(self):
        self.ext = os.path.splitext(self.name)[1]

    def run(self, *args, **kw):
        start_time = time.time()
        if not self.output_data.is_cached():
            self.set_output_data()
        self.elapsed = time.time() - start_time

class InitialVirtualArtifact(InitialArtifact):
    """
    The artifact class representing an initial virtual artifact, basically just holds the contents specified in config.
    """
    def get_contents(self):
        contents = self.args.get('contents')
        if not contents and not isinstance(contents, dict):
            msg = "No contents found for virtual file '%s'.\n" % self.key
            msg += inspect.cleandoc("""If you didn't mean to request a virtual file of this name,
            and want dexy to look for only real files, you need a wildcard character
            in the file name. Otherwise either assign contents to the virtual file
            or remove the entry from your config file.""")
            raise dexy.exceptions.UserFeedback(msg)
        return contents

    def get_contents_hash(self):
        if self.args.get('contentshash'):
            return self.args['contentshash']
        else:
            contents = self.get_contents()
            return hashlib.md5(str(contents)).hexdigest()

    def data_class_alias(self):
        data_class_alias = self.args.get('data-class-alias')

        if data_class_alias:
            return data_class_alias
        else:
            contents = self.get_contents()
            if isinstance(contents, OrderedDict):
                return 'sectioned'
            elif isinstance(contents, dict):
                return 'keyvalue'
            else:
                return 'generic'

    def set_metadata_attrs(self):
        if self.args.get('dirty'):
            self.metadata.dirty = time.time()

        self.metadata.key = self.key
        self.metadata.contentshash = self.get_contents_hash()

    def set_output_data(self):
        self.output_data.set_data(self.get_contents())

class FilterArtifact(Artifact):
    """
    Artifact class which applies filter processing to input from previous step.
    """
    def data_class_alias(self):
        if self.filter_class.PRESERVE_PRIOR_DATA_CLASS:
            return self.input_data.__class__.ALIASES[0]
        else:
            return self.filter_class.data_class_alias(self.ext)

    def calculate_canonical_name(self):
        from_args = self.canonical_name_from_args()
        if from_args:
            return from_args % self.args
        else:
            return self.filter_instance.calculate_canonical_name()

    def setup_filter_instance(self):
        self.filter_instance = self.filter_class()
        self.filter_instance.artifact = self
        self.filter_instance.log = self.log

    def run(self, *args, **kw):
        start_time = time.time()
        self.log.debug("Running %s" % self.key_with_class())

        is_cached = self.output_data.is_cached()
        cache_ok = False

        if is_cached:
            self.log.debug("Output is cached under %s, reconstituting..." % self.hashstring)
            cache_ok = self.load_cached_args()
            if cache_ok:
                self.reconstitute_cached_children()
                self.content_source = 'cached'

        if (not is_cached) or (not cache_ok):
            self.log.debug("Output is not cached under %s, running..." % self.hashstring)
            self.filter_instance.process()
            if not self.output_data.is_cached():
                if self.filter_instance.REQUIRE_OUTPUT:
                    raise dexy.exceptions.NoFilterOutput("No output file after filter ran: %s" % self.key)
            self.content_source = 'generated'

        self.elapsed = time.time() - start_time

    def load_cached_args(self):
        """
        Loads args from cache in case they have changed, returns False if cache is corrupt.
        """
        rows = self.wrapper.db.task_from_previous_batch(self.hashstring)

        if len(rows) == 0:
            rows = self.wrapper.db.task_from_current_batch(self.hashstring)

        if len(rows) == 0:
            raise dexy.exceptions.InternalDexyProblem("No rows found for %s in current or previous batch." % self.hashstring)

        raw_args = rows[0]['args']

        if not raw_args:
            return False
        else:
            for k, v in json.loads(raw_args).iteritems():
                self.log.debug("updating arg %s of %s with stored value %s" % (k, self.key, v))
                self.args[k] = v
            return True

    def reconstitute_cached_children(self):
        """
        Look for artifacts which were created as side effects of this filter
        running, re-run these docs (which should be present in cache).
        """
        rows = self.wrapper.db.get_child_hashes_in_previous_batch(self.hashstring)
        for row in rows:
            self.log.debug("Reconstituting %s from database and cache" % row['doc_key'])
            if 'Initial' in row['class_name']:
                doc_args = json.loads(row['args'])
                doc = dexy.doc.Doc(row['doc_key'], **doc_args)
                self.add_doc(doc)

                new_calc_hashstring = doc.children[0].hashstring
                db_hashstring = row['hashstring']
                msg = "unexpected calculated hashstring '%s' for %s, expected '%s'" % (new_calc_hashstring, doc.children[0].key, db_hashstring)
                assert new_calc_hashstring == db_hashstring, msg

    def set_metadata_attrs(self):
        self.metadata.dexy_version = DEXY_VERSION
        self.metadata.ext = self.ext
        self.metadata.key = self.key
        self.metadata.next_filter_name = self.next_filter_name

        # hash of the artifact supplying input to this
        self.metadata.prior_hash = self.prior.hashstring

        # args passed to this document by user
        strargs = []
        skip_arg_keys = ['wrapper']
        for k in sorted(self.args):
            if not k in skip_arg_keys:
                v = str(self.args[k])
                strargs.append("%s: %s" % (k, v))
        self.metadata.argstr = ", ".join(strargs)

        # version of external software being run, if any
        if hasattr(self.filter_class, 'version'):
            version = self.filter_class.version()
            self.metadata.software_version = version

    def add_doc(self, doc):
        if doc.state == 'complete':
            raise Exception("Already complete!")

        self.log.debug("adding additional doc %s (created by %s)" % (doc.key, self.key))
        doc.created_by_doc = self.hashstring
        doc.created_by_doc_key = self.doc.key_with_class()
        doc.wrapper = self.wrapper
        doc.canon = True

        node = dexy.node.Node(doc.key)
        node.children = [doc]
        doc.node = node

        for task in (doc,):
            for t in task:
                t()
        for task in (doc,):
            for t in task:
                t()
        for task in (doc,):
            for t in task:
                t()

        self.doc.node.inputs.append(node)

    def set_extension(self):
        this_filter_outputs = self.filter_class.OUTPUT_EXTENSIONS
        this_filter_accepts = self.filter_class.INPUT_EXTENSIONS

        # Check that we can handle input extension
        if set([self.prior.ext, ".*"]).isdisjoint(set(this_filter_accepts)):
            msg = "Filter '%s' in '%s' can't handle file extension %s, supported extensions are %s"
            params = (self.filter_alias, self.key, self.prior.ext, ", ".join(this_filter_accepts))
            raise dexy.exceptions.UserFeedback(msg % params)

        # Figure out output extension
        ext = self.filter_args().get('ext')
        if ext:
            # User has specified desired extension
            if not ext.startswith('.'):
                ext = '.%s' % ext

            # Make sure it's a valid one
            if (not ext in this_filter_outputs) and (not ".*" in this_filter_outputs):
                msg = "You have requested file extension %s in %s but filter %s can't generate that."
                raise dexy.exceptions.UserFeedback(msg % (ext, self.key, self.filter_alias))

            self.ext = ext

        elif ".*" in this_filter_outputs:
            self.ext = self.prior.ext

        else:
            # User has not specified desired extension, and we don't output wildcards,
            # figure out extension based on next filter in sequence, if any.
            if self.next_filter_class:
                next_filter_accepts = self.next_filter_class.INPUT_EXTENSIONS

                if ".*" in next_filter_accepts:
                    self.ext = this_filter_outputs[0]
                else:
                    if set(this_filter_outputs).isdisjoint(set(next_filter_accepts)):
                        msg = "Filter %s can't go after filter %s, no file extensions in common."
                        raise dexy.exceptions.UserFeedback(msg % (self.next_filter_alias, self.filter_alias))

                    for e in this_filter_outputs:
                        if e in next_filter_accepts:
                            self.ext = e

                    if not self.ext:
                        msg = "no file extension found but checked already for disjointed, should not be here"
                        raise dexy.exceptions.InternalDexyProblem(msg)
            else:
                self.ext = this_filter_outputs[0]

    def filter_args(self):
        """
        Arguments that are passed by the user which are specifically for the
        filter that will be run by this artifact.
        """
        return self.args.get(self.filter_alias, {})



from dexy.tests.utils import assert_output
from dexy.tests.utils import assert_in_output
from dexy.tests.utils import wrap
from dexy.node import DocNode

RST = """
* a bullet point using "*"

  - a sub-list using "-"

    + yet another sub-list

  - another item
"""

def test_rst2odt():
    with wrap() as wrapper:
        node = DocNode("example.txt|rst2odt",
                contents=RST,
                wrapper=wrapper)
        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().filesize() > 8000

def test_rst2xml():
    assert_in_output('rst2xml', RST, """<list_item><paragraph>a sub-list using "-"</paragraph><bullet_list bullet="+"><list_item>""")

def test_rst2latex():
    assert_in_output('rst2latex', RST, "\item a bullet point using ``*''")
    assert_in_output('rst2latex', RST, "\\begin{document}")

def test_rst2html():
    assert_in_output('rst2html', RST, "<html xmlns")
    assert_in_output('rst2html', RST, "<li>a bullet point using &quot;*&quot;<ul>")

def test_rest_to_tex():
    with wrap() as wrapper:
        node = DocNode("example.txt|rst",
                contents=RST,
                rst={"ext" : ".tex"},
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().as_text() == """\
%
\\begin{itemize}

\item a bullet point using ``*''
%
\\begin{itemize}

\item a sub-list using ``-''
%
\\begin{itemize}

\item yet another sub-list

\end{itemize}

\item another item

\end{itemize}

\end{itemize}
"""

def test_rest_to_html():
    expected = """\
<ul class="simple">
<li>a bullet point using &quot;*&quot;<ul>
<li>a sub-list using &quot;-&quot;<ul>
<li>yet another sub-list</li>
</ul>
</li>
<li>another item</li>
</ul>
</li>
</ul>
"""

    assert_output('rst', RST, expected)



from dexy.parser import AbstractSyntaxTree
from dexy.tests.utils import wrap

def test_ast():
    with wrap() as wrapper:
        ast = AbstractSyntaxTree()
        ast.wrapper = wrapper

        ast.add_task_info("abc.txt", foo='bar')
        ast.add_dependency("abc.txt", "def.txt")
        ast.add_task_info("def.txt", foo='baz')

        assert ast.tree == ['doc:abc.txt']
        assert ast.task_kwargs('doc:abc.txt')['foo'] == 'bar'
        assert ast.task_kwargs('doc:def.txt')['foo'] == 'baz'



"""
DexyFilters written to be examples of how to write filters.
"""
from dexy.common import OrderedDict
from dexy.doc import Doc
from dexy.filter import DexyFilter
import os

class Example(DexyFilter):
    """
    Base class for example filters intended to show filter development features but not be actual usable filters.
    """
    ALIASES = []
    NODOC = True

class KeyValueExample(Example):
    """
    Example of storing key value data.
    """
    ALIASES = ['keyvalueexample']
    OUTPUT_DATA_TYPE = 'keyvalue'
    OUTPUT_EXTENSIONS = ['.sqlite3', '.json']

    def process(self):
        self.output().append("foo", "bar")
        self.output().save()

class AccessOtherDocuments(Example):
    """
    Example of accessing other documents.
    """
    ALIASES = ["others"]

    def process_text(self, input_text):
        info = []
        info.append("Here is a list of previous docs in this tree (not including %s)." % self.artifact.key)

        ### @export "access-other-docs-iterate"
        for doc in self.processed():
            assert isinstance(doc, Doc)

            ### @export "access-other-docs-lens"
            n_children = len(doc.children)
            n_inputs = len(doc.inputs)

            ### @export "access-other-docs-output-length"
            if doc.output().has_data():
                length = len(doc.output().data())
            else:
                length = len(doc.output().ordered_dict())

            ### @export "access-other-docs-finish"
            info.append("%s (%s children, %s inputs, length %s)" % (doc.key, n_children, n_inputs, length))
        s = "%s        " % os.linesep
        return s.join(info)
        ### @end

class AddNewDocument(Example):
    """
    A filter which adds an extra document to the tree.
    """
    ALIASES = ['newdoc']

    def process_text(self, input_text):
        self.add_doc("newfile.txt|processtext", "newfile")
        return "we added a new file"

class ConvertDict(Example):
    """
    Returns an ordered dict with a single element.
    """
    ALIASES = ['dict']

    def process_text_to_dict(self, input_text):
        ordered_dict = OrderedDict()
        ordered_dict['1'] = input_text
        return ordered_dict

class ExampleProcessTextMethod(Example):
    """
    Uses process_text method
    """
    ALIASES = ['processtext']

    def process_text(self, input_text):
        return "Dexy processed the text '%s'" % input_text

class ExampleProcessDictMethod(Example):
    """
    Uses process_dict method
    """
    ALIASES = ['processdict']

    def process_dict(self, input_dict):
        output_dict = OrderedDict()
        for k, v in input_dict.iteritems():
            output_dict[k] = "Dexy processed the text '%s'" % v
        return output_dict

class ExampleProcessTextToDictMethod(Example):
    """
    Uses process_text_do_dict method
    """
    ALIASES = ['processtexttodict']

    def process_text_to_dict(self, input_text):
        output_dict = OrderedDict()
        output_dict['1'] = "Dexy processed the text '%s'" % input_text
        return output_dict

class ExampleProcessMethod(Example):
    """
    A filter implementing a process method which stores raw data.
    """
    ALIASES = ['process']

    def process(self):
        output = "Dexy processed the text '%s'" % self.input().data()
        self.output().set_data(output)

class ExampleProcessMethodManualWrite(Example):
    """
    A filter implementing a process method which stores raw data by writing directly to the output file.
    """
    ALIASES = ['processmanual']

    def process(self):
        input_data = self.input().data()
        output = "Dexy processed the text '%s'" % input_data
        with open(self.output_filepath(), "wb") as f:
            f.write(output)

class ExampleProcessWithDictMethod(Example):
    """
    A filter implementing a process method which uses OrderedDict to store sectional data.
    """
    ALIASES = ['processwithdict']
    OUTPUT_DATA_TYPE = 'sectioned'

    def process(self):
        input_data = self.input().data()
        output_dict = OrderedDict()
        output_dict['1'] = "Dexy processed the text '%s'" % input_data
        self.output().set_data(output_dict)

class AbcExtension(Example):
    """
    Only outputs extension .abc
    """
    OUTPUT_EXTENSIONS = [".abc"]
    ALIASES = ['outputabc']

    def process_text(self, input_text):
        return "Dexy processed the text '%s'" % input_text

class ExampleFilterArgs(Example):
    """
    Prints out the args it receives.
    """
    ALIASES = ['filterargs']

    def process_text(self, input_text):
        result = ["Here are the arguments you passed:"]
        for k in sorted(self.args()):
            v = self.args()[k]
            result.append("%s: %s" % (k, v))
        return os.linesep.join(result)



from datetime import datetime
from dexy.plugin import PluginMeta
from dexy.version import DEXY_VERSION
from pygments.styles import get_all_styles
import calendar
import dexy.artifact
import dexy.commands
import dexy.data
import dexy.exceptions
import json
import os
import pygments
import pygments.formatters
import re

class TemplatePlugin():
    __metaclass__ = PluginMeta

    @classmethod
    def is_active(klass):
        return True

    def __init__(self, filter_instance):
        self.filter_instance = filter_instance
        self.log = self.filter_instance.artifact.log

    def run(self):
        return {}

try:
    from bs4 import BeautifulSoup
    BS4_AVAILABLE = True
except ImportError:
    BS4_AVAILABLE = False

class PrettyPrintHtml(TemplatePlugin):
    """
    Uses BeautifulSoup 4 to prettify HTML.
    """
    @classmethod
    def is_active(klass):
        return BS4_AVAILABLE

    @classmethod
    def prettify_html(klass, html):
        soup = BeautifulSoup(unicode(html))
        return soup.prettify()

class JavadocToRst(TemplatePlugin):
    """
    Exposes javadoc2rst command which strips HTML tags from javadoc comments.
    """

    ESCAPE = ['\\']
    REMOVE = ['<p>', '<P>']

    @classmethod
    def javadoc2rst(klass, javadoc):
        for x in klass.REMOVE:
            javadoc = javadoc.replace(x, '\n')
        for x in klass.ESCAPE:
            javadoc = javadoc.replace(x, "\\%s" % x)
        return javadoc

class PrettyPrintJson(TemplatePlugin):
    """
    Exposes ppjson command.
    """
    ALIASES = ['ppjson']

    def ppjson(self, json_string):
        return json.dumps(json.loads(json_string), sort_keys = True, indent = 4)

    def run(self):
        return {
            'ppjson' : self.ppjson
         }

class JinjaFilters(TemplatePlugin):
    """
    Custom jinja filters which can be accessed using the | fn() syntax in jinja templates.
    """
    @classmethod
    def indent(klass, s, width=4, indentfirst=False):
        """
        Replacement for jinja's indent method to ensure unicode() is called prior to splitlines().
        """
        output = []
        for i, line in enumerate(unicode(s).splitlines()):
            if indentfirst or i > 0:
                output.append("%s%s" % (' ' * width, line))
            else:
                output.append(line)
        return "\n".join(output)

    @classmethod
    def head(klass, s, n=15):
        """
        Returns the first n lines of input string.
        """
        lines = unicode(s).splitlines()[0:n]
        return "\n".join(lines)

class RstCode(TemplatePlugin):
    """
    Indents code 4 spaces and wraps in .. code:: directive.
    """
    @classmethod
    def rstcode(klass, text, language='python'):
        output = [
            ".. code:: %s" % language,
            '   :class: highlight',
            '']
        for line in unicode(text).splitlines():
            output.append("    %s" % line)
        output.append('')
        return os.linesep.join(output)

class PythonDatetime(TemplatePlugin):
    """
    Exposes python datetime and calendar functions.
    """
    ALIASES = ['datetime', 'calendar']
    def run(self):
        today = datetime.today()
        month = today.month
        year = today.year
        cal = calendar.Calendar()
        caldates = list(cal.itermonthdates(year, month))

        return {
            "datetime" : datetime,
            "calendar" : calendar,
            "caldates" : caldates,
            "cal" : cal,
            "today" : today,
            "month" : month,
            "year" : year
            }

class DexyVersion(TemplatePlugin):
    """
    Exposes the current dexy version
    """
    ALIASES = ['dexyversion']
    def run(self):
        return { "DEXY_VERSION" : DEXY_VERSION }

class SimpleJson(TemplatePlugin):
    """
    Exposes the json module.
    """
    ALIASES = ['json']
    def run(self):
        return { 'json' : json }

class RegularExpressions(TemplatePlugin):
    """
    Exposes re_match and re_search.
    """
    ALIASES = ['regex']
    def run(self):
        return { 're_match' : re.match, 're_search' : re.search}

class PythonBuiltins(TemplatePlugin):
    """
    Exposes python builtins.
    """
    ALIASES = ['builtins']
    # Intended to be all builtins that make sense to run within a document.
    PYTHON_BUILTINS = [abs, all, any, bin, bool, bytearray, callable, chr,
        cmp, complex, dict, dir, divmod, enumerate, filter, float, format, hex,
        id, int, isinstance, issubclass, iter, len, list, locals, long, map,
        hasattr, max, min, oct, ord, pow, range, reduce, repr, reversed, round,
        set, slice, sorted, str, sum, tuple, xrange, unicode, zip]

    def run(self):
        return dict((f.__name__, f) for f in self.PYTHON_BUILTINS)

class PygmentsStylesheet(TemplatePlugin):
    """
    Generates pygments stylesheets.
    """
    ALIASES = ['pygments']

    # TODO figure out default fmt based on document ext
    @classmethod
    def highlight(klass, text, lexer_name, fmt = 'html', noclasses = False, lineanchors = 'l'):
        if text:
            formatter_options = { "lineanchors" : lineanchors, "noclasses" : noclasses }
            lexer = pygments.lexers.get_lexer_by_name(lexer_name)
            formatter = pygments.formatters.get_formatter_by_name(fmt, **formatter_options)
            return pygments.highlight(text, lexer, formatter)

    def run(self):
        pygments_stylesheets = {}
        if self.filter_instance.artifact.args.has_key('pygments'):
            formatter_args = self.filter_instance.artifact.args['pygments']
        else:
            formatter_args = {}

        for style_name in get_all_styles():
            for formatter_class in [pygments.formatters.LatexFormatter, pygments.formatters.HtmlFormatter]:
                formatter_args['style'] = style_name
                pygments_formatter = formatter_class(**formatter_args)
                style_info = pygments_formatter.get_style_defs()

                for fn in pygments_formatter.filenames:
                    ext = fn.split(".")[1]
                    if ext == 'htm':
                        ext = 'css' # swap the more intuitive '.css' for the unlikely '.htm'
                    key = "%s.%s" % (style_name, ext)
                    pygments_stylesheets[key] = style_info
        return {'pygments' : pygments_stylesheets, 'highlight' : self.highlight }

class Subdirectories(TemplatePlugin):
    """
    Show subdirectories under this document.
    """
    ALIASES = ['subdirectories']
    def run(self):
        # The directory containing the document to be processed.
        doc_dir = os.path.dirname(self.filter_instance.output().name)

        # Get a list of subdirectories under this document's directory.
        subdirectories = [d for d in sorted(os.listdir(os.path.join(os.curdir, doc_dir))) if os.path.isdir(os.path.join(os.curdir, doc_dir, d))]
        return {'subdirectories' : subdirectories}

class Variables(TemplatePlugin):
    """
    Allow users to set variables in document args which will be available to an individual document.
    """
    ALIASES = ['variables']
    def run(self):
        variables = {}
        variables.update(self.filter_instance.arg_value('variables', {}))
        variables.update(self.filter_instance.arg_value('vars', {}))
        return variables

class Globals(TemplatePlugin):
    """
    Makes available the global variables specified on the dexy command line
    using the --globals option
    """
    ALIASES = ['globals']
    def run(self):
        raw_globals = self.filter_instance.artifact.wrapper.globals
        env = {}
        for kvpair in raw_globals.split(","):
            if "=" in kvpair:
                k, v = kvpair.split("=")
                env[k] = v
        return env

class Inputs(TemplatePlugin):
    """
    Populates the 'd' object.
    """
    ALIASES = ['inputs']

    def input_tasks(self):
        for doc in self.filter_instance.processed():
            yield doc.final_artifact

    def a(self, relative_ref):
        return self.map_relative_refs[relative_ref]

    def run(self):
        self.map_relative_refs = {}

        for task in self.input_tasks():
            for ref in task.output_data.relative_refs(self.filter_instance.output().name):
                self.map_relative_refs[ref] = task.output_data

        return {
            'a' : self.a,
            'args' : self.filter_instance.artifact.args,
            'd' : D(self.filter_instance.artifact, self.map_relative_refs),
            'f' : self.filter_instance,
            's' : self.filter_instance.output(),
            'w' : self.filter_instance.artifact.wrapper
            }

class D(object):
    def __init__(self, artifact, map_relative_refs):
        self._artifact = artifact
        self._map_relative_refs = map_relative_refs

    def __getitem__(self, relative_ref):
        if self._map_relative_refs.has_key(relative_ref):
            return self._map_relative_refs[relative_ref]
        else:
            raise dexy.exceptions.UserFeedback("There is no document named %s available to %s" % (relative_ref, self._artifact.key))



from dexy.common import OrderedDict
from dexy.node import DocNode
from dexy.tests.utils import assert_output
from dexy.tests.utils import wrap
import os

def test_header_footer_filters():
    with wrap() as wrapper:
        os.makedirs('subdir/subsubdir')
        node = DocNode("subdir/file.txt|hd|ft",
                contents="These are main contents.",
                inputs = [
                    DocNode("_header.txt", wrapper=wrapper, contents="This is a header in parent dir."),
                    DocNode("subdir/_header.txt|jinja", wrapper=wrapper, contents="This is a header."),
                    DocNode("subdir/_footer.txt|jinja", wrapper=wrapper, contents="This is a footer."),
                    DocNode("subdir/subsubdir/_header.txt", wrapper=wrapper, contents="This is a header in a subdirectory.")
                    ],
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().data() == "This is a header.\nThese are main contents.\nThis is a footer."

def test_join_filter():
    contents = OrderedDict()
    contents['1'] = "section one"
    contents['2'] = "section two"

    assert_output("join", contents, "section one\nsection two")

def test_head_filter():
    assert_output("head", "1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n", "1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n")

def test_word_wrap_filter():
    with wrap() as wrapper:
        node = DocNode("example.txt|wrap", contents="this is a line of text", wrap={"width" : 5}, wrapper=wrapper)
        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().data() == "this\nis a\nline\nof\ntext"

def test_lines_filter():
    expected = OrderedDict()
    expected['1'] = "line one"
    expected['2'] = "line two"

    assert_output("lines", "line one\nline two", expected)

def test_ppjson_filter():
    assert_output(
            "ppjson",
            '{"foo" :123, "bar" :456}',
            """{\n    "bar": 456, \n    "foo": 123\n}"""
            )

def test_start_space_filter():
    o = OrderedDict()
    o['1'] = " abc\n def"
    assert_output("startspace", "abc\ndef", o)

def test_tags_filter():
    with wrap() as wrapper:
        node = DocNode("example.txt|tags", contents="<p>the text</p>", tags={"tags" : ["html", "body"]}, wrapper=wrapper)
        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().data() == "<html><body>\n<p>the text</p>\n</body></html>"




from dexy.common import OrderedDict
from dexy.doc import Doc
from dexy.tests.utils import assert_in_output
from dexy.tests.utils import assert_output
from dexy.tests.utils import assert_output_cached
from dexy.tests.utils import wrap
from nose.tools import raises
import dexy.exceptions

def test_pyg4rst():
    o = OrderedDict()
    o['1'] = ".. code:: python\n\n  print 'hello'"
    assert_output("pyg4rst", "print 'hello'", o, ext=".py")

def test_html():
    assert_in_output("pyg|h", "print 'hello'", """<div class="highlight">""")

def test_png():
    assert_output_cached("pyg|pn", "print 'hello'")

def test_jpg():
    assert_output_cached("pyg|jn", "print 'hello'")

def test_gif():
    assert_output_cached("pyg|gn", "print 'hello'")

@raises(dexy.exceptions.UserFeedback)
def test_pyg4rst_bad_file_extension():
    with wrap() as wrapper:
        doc = Doc("hello.xyz|pyg4rst", wrapper=wrapper, contents=" ")
        wrapper.run_docs(doc)

@raises(dexy.exceptions.UserFeedback)
def test_pygments_bad_file_extension():
    with wrap() as wrapper:
        doc = Doc("hello.xyz|pyg", wrapper=wrapper, contents=" ")
        wrapper.run_docs(doc)



from dexy.tests.utils import wrap
from dexy.plugins.parsers import YamlFileParser

YAML = """foo:
    - bar
    - baz

foob:
    - foobar

xyz:
    - abc
    - def
"""

def run_yaml_with_target(target):
    with wrap() as wrapper:
        parser = YamlFileParser()
        parser.wrapper = wrapper
        parser.parse(YAML)
        assert len(wrapper.batch.tree) == 3

        wrapper.batch.run(target)
        yield wrapper.batch

def test_run_target_foo():
    for batch in run_yaml_with_target("foo"):
        # foo and children have been run
        assert batch.lookup_table['BundleNode:foo'].state == 'complete'
        assert batch.lookup_table['BundleNode:bar'].state == 'complete'
        assert batch.lookup_table['BundleNode:baz'].state == 'complete'

        # foob and children have not been run
        assert batch.lookup_table['BundleNode:foob'].state == 'new'
        assert batch.lookup_table['BundleNode:foobar'].state == 'new'

def test_run_target_fo():
    for batch in run_yaml_with_target("fo"):
        # foo and children have been run
        assert batch.lookup_table['BundleNode:foo'].state == 'complete'
        assert batch.lookup_table['BundleNode:bar'].state == 'complete'
        assert batch.lookup_table['BundleNode:baz'].state == 'complete'

        # foob and children have been run
        assert batch.lookup_table['BundleNode:foob'].state == 'complete'
        assert batch.lookup_table['BundleNode:foobar'].state == 'complete'

def test_run_target_bar():
    for batch in run_yaml_with_target("bar"):
        assert batch.lookup_table['BundleNode:foo'].state == 'new'
        assert batch.lookup_table['BundleNode:bar'].state == 'complete'
        assert batch.lookup_table['BundleNode:baz'].state == 'new'
        assert batch.lookup_table['BundleNode:foob'].state == 'new'
        assert batch.lookup_table['BundleNode:foobar'].state == 'new'

def test_run_target_ba():
    for batch in run_yaml_with_target("ba"):
        assert batch.lookup_table['BundleNode:foo'].state == 'new'
        assert batch.lookup_table['BundleNode:bar'].state == 'complete'
        assert batch.lookup_table['BundleNode:baz'].state == 'complete'
        assert batch.lookup_table['BundleNode:foob'].state == 'new'
        assert batch.lookup_table['BundleNode:foobar'].state == 'new'



from dexy.tests.utils import runfilter
import json

BEOWULF = """BEOWULF

PRELUDE OF THE FOUNDER OF THE DANISH HOUSE

LO, praise of the prowess of people-kings
of spear-armed Danes, in days long sped,
we have heard, and what honor the athelings won!
Oft Scyld the Scefing from squadroned foes,
"""

def test_nltk_filter():
    with runfilter("nltk", BEOWULF) as doc:
        output = json.loads(doc.output().as_text())

        assert "wc" in output.keys()
        assert "collocations" in output.keys()
        assert "pos-tags" in output.keys()

        assert output['wc'] == 45
        assert ["DANISH", "HOUSE"] in output['collocations']



from dexy.exceptions import UnexpectedState, InvalidStateTransition, CircularDependency
from dexy.tests.utils import divert_stdout
from dexy.tests.utils import wrap
from nose.tools import raises
import dexy.task

def test_init():
    task = dexy.task.Task("key")
    assert isinstance(task.args, dict)
    assert isinstance(task.children, list)
    assert task.key == "key"

    assert hasattr(task, 'pre')
    assert hasattr(task, 'post')
    assert hasattr(task.pre, "__call__")
    assert hasattr(task.post, "__call__")

def test_kwargs():
    task = dexy.task.Task("key", foo="bar", abc="def")
    assert task.args['foo'] == 'bar'
    assert task.args['abc'] == 'def'

@raises(UnexpectedState)
def test_invalid_state():
    task = dexy.task.Task("key")
    task.state = "invalid"
    for t in task:
        t()

@raises(InvalidStateTransition)
def test_invalid_transition():
    task = dexy.task.Task('key')
    task.transition('complete')

def test_set_log():
    task = dexy.task.Task("key")
    assert not hasattr(task, 'log')
    task.set_log()
    assert hasattr(task, 'log')
    assert hasattr(task, 'logstream')
    task.log.debug("please write me to log")
    assert "please write me to log" in task.logstream.getvalue()

def test_circular():
    with wrap() as wrapper:
        wrapper.setup_batch()

        d1 = dexy.task.Task("1",wrapper=wrapper)
        d2 = dexy.task.Task("2",wrapper=wrapper)

        d1.children.append(d2)
        d2.children.append(d1)

        try:
            for t in d1:
                t()
            assert False
        except CircularDependency:
            assert True

@raises(CircularDependency)
def test_circular_4_docs():
    with wrap() as wrapper:
        wrapper.setup_batch()

        d1 = dexy.task.Task("1", wrapper=wrapper)
        d2 = dexy.task.Task("2", wrapper=wrapper)
        d3 = dexy.task.Task("3", wrapper=wrapper)
        d4 = dexy.task.Task("4", wrapper=wrapper)

        d1.children.append(d2)
        d2.children.append(d3)
        d3.children.append(d4)
        d4.children.append(d1)

        for t in d1:
            t()

class SubclassTask(dexy.task.Task):
    """
    for testing
    """
    def pre(self, *args, **kw):
        self.hashstring = ''
        print "pre '%s'" % self.key,

    def run(self, *args, **kw):
        print "run '%s'" % self.key,

    def post(self, *args, **kw):
        print "post '%s'" % self.key,

def test_run_demo_single():
    with divert_stdout() as stdout:
        with wrap() as wrapper:
            doc = SubclassTask("demo", wrapper=wrapper)

            wrapper.run_docs(doc)

            assert "pre 'demo' run 'demo' post 'demo'" == stdout.getvalue()

def test_run_demo_parent_child():
    with divert_stdout() as stdout:
        with wrap() as wrapper:
            doc = SubclassTask("parent", wrapper=wrapper)
            doc.children = [SubclassTask("child", wrapper=wrapper)]
            wrapper.run_docs(doc)
        
        assert "pre 'parent' pre 'child' run 'child' post 'child' run 'parent' post 'parent'" == stdout.getvalue()

def test_completed_children():
    with wrap() as wrapper:
        with divert_stdout() as stdout:
            grandchild_task = SubclassTask("grandchild", wrapper=wrapper)
            child_task = SubclassTask("child", wrapper=wrapper)
            parent_task = SubclassTask("parent", wrapper=wrapper)

            parent_task.children = [child_task]
            child_task.children = [grandchild_task]

            wrapper.run_docs(parent_task)

            assert stdout.getvalue() == "pre 'parent' pre 'child' pre 'grandchild' run 'grandchild' post 'grandchild' run 'child' post 'child' run 'parent' post 'parent'"



from dexy.common import OrderedDict
from dexy.filter import DexyFilter
from dexy.plugins.standard_filters import StartSpaceFilter
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.formatters import LatexFormatter
from pygments.formatters import get_formatter_for_filename
from pygments.lexers import get_lexer_by_name
from pygments.lexers import get_lexer_for_filename
from pygments.lexers.agile import PythonConsoleLexer
from pygments.lexers.agile import RubyConsoleLexer
from pygments.lexers.templates import DjangoLexer
from pygments.lexers.text import MakefileLexer
from pygments.lexers.text import TexLexer
from pygments.lexers.web import JavascriptLexer
from pygments.lexers.web import XmlLexer
import dexy.commands
import dexy.exceptions
import pygments.lexers.web
import re

class SyntaxHighlightRstFilter(DexyFilter):
    """
    Surrounds code with highlighting instructions for ReST
    """
    ALIASES = ['pyg4rst']

    def process_dict(self, input_dict):
        n = self.args().get('n', 2)
        result = OrderedDict()

        try:
            lexer = get_lexer_for_filename(self.input().storage.data_file())
        except pygments.util.ClassNotFound:
            msg = PygmentsFilter.LEXER_ERR_MSG
            raise dexy.exceptions.UserFeedback(msg % (self.input().name, self.artifact.key))

        lexer_alias = lexer.aliases[0]

        for section_name, section_text in input_dict.iteritems():
            with_spaces = StartSpaceFilter.add_spaces_at_start(section_text, n)
            result[section_name] = ".. code:: %s\n\n%s" % (lexer_alias, with_spaces)

        return result

class PygmentsFilter(DexyFilter):
    """
    Apply Pygments syntax highlighting. Image formats require PIL.
    """
    INPUT_EXTENSIONS = [".*"]
    IMAGE_OUTPUT_EXTENSIONS = ['.png', '.bmp', '.gif', '.jpg']
    MARKUP_OUTPUT_EXTENSIONS = [".html", ".tex", ".svg"] # make sure .html is first!
    OUTPUT_EXTENSIONS = MARKUP_OUTPUT_EXTENSIONS + IMAGE_OUTPUT_EXTENSIONS + ['.css', '.sty']
    ALIASES = ['pyg', 'pygments']
    FRAGMENT = True
    LEXER_ERR_MSG = "Pygments doesn't know how to syntax highlight files like '%s' (for '%s'). Either it can't be done or you need to specify the lexer manually."

    @classmethod
    def data_class_alias(klass, file_ext):
        if file_ext in klass.MARKUP_OUTPUT_EXTENSIONS:
            return 'sectioned'
        else:
            return 'generic'

    @classmethod
    def docmd_css(klass, style='default'):
        print klass.generate_css(style)

    @classmethod
    def docmd_sty(klass, style='default'):
        print klass.generate_sty(style)

    @classmethod
    def generate_css(self, style='default'):
        formatter = HtmlFormatter(style=style)
        return formatter.get_style_defs()

    @classmethod
    def generate_sty(self, style='default'):
        formatter = LatexFormatter(style=style)
        return formatter.get_style_defs()

    def calculate_canonical_name(self):
        ext = self.artifact.prior.ext
        if ext in [".css", ".sty"] and self.artifact.ext == ext:
            return self.artifact.doc.name
        else:
            return "%s%s" % (self.artifact.doc.name, self.artifact.ext)

    def create_lexer_instance(self, args):
        ext = self.artifact.prior.ext
        lexer = None
        lexer_args = {}

        # Pick out any options which are prefixed with "lexer-"
        for k in args.keys():
            m = re.match("^lexer-(.+)$", k)
            if m:
                option_name = m.groups()[0]
                option_value = args[k]
                self.log.debug("using custom lexer option for %s with value %s" % (option_name, option_value))
                del args[k]
                lexer_args[option_name] = option_value

        # Create a lexer instance.
        if args.has_key('lexer'):
            self.log.debug("custom lexer alias %s specified" % args['lexer'])
            lexer = get_lexer_by_name(args['lexer'], **lexer_args)
            del args['lexer']
        else:
            is_json_file = ext in ('.json', '.dexy') or self.output().name.endswith(".dexy")
            if ext == '.pycon':
                lexer_class = PythonConsoleLexer
            elif ext == '.rbcon':
                lexer_class = RubyConsoleLexer
            elif is_json_file and (pygments.__version__ < '1.5'):
                lexer_class = JavascriptLexer
            elif is_json_file:
                lexer_class = pygments.lexers.web.JSONLexer
            elif ext == '.Rd':
                lexer_class = TexLexer # does a passable job
            elif ext == '.svg':
                lexer_class = XmlLexer
            elif ext == '.jinja':
                lexer_class = DjangoLexer
            elif ext == '.Makefile' or (ext == '' and 'Makefile' in self.input().name):
                lexer_class = MakefileLexer
            else:
                fake_file_name = "input_text%s" % ext
                try:
                    lexer = get_lexer_for_filename(fake_file_name, **lexer_args)
                except pygments.util.ClassNotFound:
                    raise dexy.exceptions.UserFeedback(self.LEXER_ERR_MSG % (self.input().name, self.artifact.key))

            if not lexer:
                lexer = lexer_class(**lexer_args)

        self.log.debug("using lexer %s" % lexer.__class__.__name__)
        return lexer

    def create_formatter_instance(self, args):
        formatter_args = {'lineanchors' : self.output().web_safe_document_key() }

        # Python 2.6 doesn't like unicode keys as kwargs
        for k, v in args.iteritems():
            formatter_args[str(k)] = v

        self.log.debug("creating pygments formatter with args %s" % (formatter_args))
        return get_formatter_for_filename(self.output().name, **formatter_args)

    def process(self):
        if self.artifact.ext in self.IMAGE_OUTPUT_EXTENSIONS:
            try:
                import PIL
            except ImportError:
                print "python imaging library is required by pygments to create image output"
                raise dexy.exceptions.InactiveFilter('pyg', self.artifact.key)


        ext = self.artifact.prior.ext
        if ext in [".css", ".sty"] and self.artifact.ext == ext:
            self.log.debug("creating a style file in %s" % self.artifact.key)
            # Special case if we get a virtual empty file, generate style file
            if ext == '.css':
                output = self.generate_css(self.arg_value('style', 'default'))
            elif ext == '.sty':
                output = self.generate_sty(self.arg_value('style', 'default'))
            else:
                raise dexy.commands.UserFeedback("pyg filter doesn't know how to generate a stylesheet for %s extension" % ext)

            self.output().set_data(output)

        else:
            args = self.args().copy()
            lexer = self.create_lexer_instance(args)

            formatter_args = {'lineanchors' : self.output().web_safe_document_key() }

            # Python 2.6 hates unicode keys
            for k, v in args.iteritems():
                formatter_args[str(k)] = v

            if self.artifact.ext in self.IMAGE_OUTPUT_EXTENSIONS:
                # Place each section into an image.
                for k, v in self.input().as_sectioned().iteritems():
                    formatter = get_formatter_for_filename(self.output().name, **formatter_args)
                    output_for_section = highlight(v.decode("utf-8"), lexer, formatter)
                    new_doc_name = "%s--%s%s" % (self.artifact.doc.key.replace("|", "--"), k, self.artifact.ext)
                    self.add_doc(new_doc_name, output_for_section)

                # Place entire contents into main file.
                formatter = get_formatter_for_filename(self.output().name, **formatter_args)
                self.artifact.doc.canon = True
                with open(self.output_filepath(), 'wb') as f:
                    f.write(highlight(self.input().as_text(), lexer, formatter))

            else:
                formatter = get_formatter_for_filename(self.output().name, **formatter_args)
                output_dict = OrderedDict()
                for k, v in self.input().as_sectioned().iteritems():
                    output_dict[k] = highlight(v.decode("utf-8"), lexer, formatter)
                self.output().set_data(output_dict)



from dexy.node import DocNode
from dexy.plugins.process_filters import SubprocessFilter
from dexy.tests.utils import wrap
import dexy.exceptions
import os

def test_add_new_files():
    with wrap() as wrapper:
        node = DocNode("example.sh|sh",
                contents = "echo 'hello' > newfile.txt",
                sh = {
                    "add-new-files" : True,
                    "additional-doc-filters" : { '.txt' : 'markdown' }
                    },
                wrapper=wrapper)

        wrapper.run_docs(node)

        assert wrapper.batch.lookup_table['Doc:newfile.txt'].output().data() == "hello" + os.linesep
        assert wrapper.batch.lookup_table['Doc:newfile.txt|markdown'].output().data() == "<p>hello</p>"

def test_walk_working_dir():
    with wrap() as wrapper:
        node = DocNode("example.sh|sh",
                contents = "echo 'hello' > newfile.txt",
                sh = {
                    "walk-working-dir" : True,
                    },
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]

        for doc in wrapper.batch.tasks():
            if doc.key_with_class() == "Doc:example.sh-sh.txt-files":
                assert doc.output().as_sectioned()['newfile.txt'] == "hello" + os.linesep


def test_not_present_executable():
    assert 'notreal' in NotPresentExecutable.executables()
    assert not NotPresentExecutable.executable()

class NotPresentExecutable(SubprocessFilter):
    """
    notreal
    """
    EXECUTABLE = 'notreal'

def test_command_line_args():
    with wrap() as wrapper:
        node = DocNode("example.py|py",
                py={"args" : "-B"},
                wrapper=wrapper,
                contents="print 'hello'"
                )
        wrapper.run_docs(node)
        doc = node.children[0]

        assert doc.output().data() == "hello" + os.linesep

        command_used = doc.children[-1].filter_instance.command_string()
        assert command_used == """python -B "example.py"  "example.txt" """

def test_scriptargs():
    with wrap() as wrapper:
        node = DocNode("example.py|py",
                py={"scriptargs" : "--foo"},
                wrapper=wrapper,
                contents="import sys\nprint sys.argv[1]"
                )
        wrapper.run_docs(node)
        doc = node.children[0]

        assert doc.output().data() == "--foo" + os.linesep

        command_used = doc.children[-1].filter_instance.command_string()
        assert command_used == """python  "example.py" --foo "example.txt" """

def test_custom_env_in_args():
    with wrap() as wrapper:
        node = DocNode("example.py|py",
                py={"env" : {"FOO" : "bar" }},
                wrapper=wrapper,
                contents="import os\nprint os.environ['FOO']"
                )
        wrapper.run_docs(node)

        doc = node.children[0]

        assert doc.output().data() == "bar" + os.linesep

def test_nonzero_exit():
    with wrap() as wrapper:
        node = DocNode("example.py|py",
                wrapper=wrapper,
                contents="import sys\nsys.exit(1)"
                )
        try:
            wrapper.run_docs(node)
            assert False, "should raise error"
        except dexy.exceptions.UserFeedback:
            assert True

def test_ignore_nonzero_exit():
    with wrap() as wrapper:
        wrapper.ignore_nonzero_exit = True
        node = DocNode("example.py|py",
                wrapper=wrapper,
                contents="import sys\nsys.exit(1)"
                )
        wrapper.run_docs(node)
        assert True # no NonzeroExit was raised...



from dexy.doc import Doc
from dexy.tests.utils import wrap

def test_filter_args():
    with wrap() as wrapper:
        doc = Doc(
                "hello.txt|filterargs",
                contents="hello",
                foo="bar",
                filterargs={"abc" : 123, "foo" : "baz" },
                wrapper=wrapper)

        wrapper.run_docs(doc)

        result = doc.output().data()

        assert "Here are the arguments you passed:" in result
        assert "abc: 123" in result
        assert "foo: baz" in result
        assert not "foo: bar" in result




from dexy.version import DEXY_VERSION
import inspect
import platform

class UserFeedback(Exception):
    pass

class NoFilterOutput(UserFeedback):
    pass

class InactiveFilter(UserFeedback):
    def __init__(self, filter_alias, doc_key):
        self.message = inspect.cleandoc("""
        You are trying to use a filter '%s' which isn't active.
        Some additional software may need to be installed first.
        Either stop using this filter or install the software to continue your dexy run.
        '%s' is being used in '%s'
        """ % (filter_alias, filter_alias, doc_key))

class CircularDependency(UserFeedback):
    pass

class BlankAlias(UserFeedback):
    pass

class InvalidStateTransition(Exception):
    pass

class UnexpectedState(Exception):
    pass

class InternalDexyProblem(Exception):
    def __init__(self, message):
        self.message = inspect.cleandoc("""
        Oops! You may have found a bug in Dexy.
        The developer would really appreciate if you copy and paste this entire message
        and the Traceback above it into an email and send to info@dexy.it
        Your version of Dexy is %s
        Your platform is %s""" % (DEXY_VERSION, platform.system()))
        self.message += "\n"
        self.message += message

    def __str__(self):
        return self.message



from dexy.filter import DexyFilter
import dexy.exceptions
import logging
import json

try:
    import markdown
    AVAILABLE = True
except ImportError as e:
    AVAILABLE = False

class MarkdownFilter(DexyFilter):
    """
    Runs a Markdown processor to convert markdown to HTML.

    Markdown extensions can be enabled in your config:
    http://packages.python.org/Markdown/extensions/index.html
    """
    INPUT_EXTENSIONS = ['.*']
    OUTPUT_EXTENSIONS = ['.html']
    ALIASES = ['markdown']
    DEFAULT_EXTENSIONS = {'toc' : {}}

    @classmethod
    def is_active(klass):
        return AVAILABLE

    def process_text(self, input_text):
        markdown_logger = logging.getLogger('MARKDOWN')
        markdown_logger.addHandler(self.log.handlers[-1])

        if len(self.args()) > 0:
            extensions = self.args().keys()
            extension_configs = self.args()
        else:
            extensions = self.DEFAULT_EXTENSIONS.keys()
            extension_configs = self.DEFAULT_EXTENSIONS

        dbg = "Initializing Markdown with extensions: %s and extension configs: %s"
        self.log.debug(dbg % (json.dumps(extensions), json.dumps(extension_configs)))

        try:
            md = markdown.Markdown(
                    extensions=extensions,
                    extension_configs=extension_configs)
        except ValueError as e:
            self.log.debug(e.message)
            if "markdown.Extension" in e.message:
                raise dexy.exceptions.UserFeedback("There's a problem with the markdown extensions you specified.")
            else:
                raise e
        except KeyError as e:
            raise dexy.exceptions.UserFeedback("Couldn't find a markdown extension option matching '%s'" % e.message)

        return md.convert(input_text)



from dexy.reporter import Reporter
import os

class OutputReporter(Reporter):
    """
    Creates canonical dexy output with files given short filenames.
    """
    ALIASES = ['output']
    REPORTS_DIR = 'output'

    def write_canonical_doc(self, doc):
        fp = os.path.join(self.REPORTS_DIR, doc.output().name)

        parent_dir = os.path.dirname(fp)
        if not os.path.exists(parent_dir):
            os.makedirs(os.path.dirname(fp))

        self.log.debug("  writing %s to %s" % (doc.key, fp))
        if os.path.exists(fp):
            print "WARNING %s is overwriting file %s. Already written to by:" % (doc.key, fp)
            for tup in self.keys_to_outfiles:
                k, v = tup
                if v == fp:
                    print "    %s" % k
        self.keys_to_outfiles.append((doc.key, fp))

        doc.output().output_to_file(fp)

    def run(self, wrapper):
        self.wrapper=wrapper
        self.set_log()
        self.keys_to_outfiles = []

        self.create_reports_dir()
        for doc in wrapper.batch.docs():
            if doc.canon:
                self.write_canonical_doc(doc)

class LongOutputReporter(Reporter):
    """
    Creates complete dexy output with files given long, unique filenames.
    """
    ALIASES = ['long']
    REPORTS_DIR = 'output-long'

    def run(self, wrapper):
        self.wrapper=wrapper
        self.set_log()
        self.create_reports_dir()
        for doc in wrapper.batch.docs():
            fp = os.path.join(self.REPORTS_DIR, doc.output().long_name())

            parent_dir = os.path.dirname(fp)
            if not os.path.exists(parent_dir):
                os.makedirs(os.path.dirname(fp))

            self.log.debug("  writing %s to %s" % (doc.key, fp))
            doc.output().output_to_file(fp)



from dexy.plugins.parsers import YamlFileParser
from dexy.tests.utils import wrap
from dexy.tests.utils import tempdir
import os
import dexy.wrapper

def test_subdir_config_with_bundle():
    with tempdir():

        with open("dexy.yaml", "w") as f:
            f.write("""
            foo:
                - .txt
            """)

        os.makedirs("abc/def")
        with open("abc/def/dexy.yaml", "w") as f:
            f.write("""
            bar:
                - .py
            """)

        with open("abc/def/hello.py", "w") as f:
            f.write("print 'hello'")

        wrapper = dexy.wrapper.Wrapper()
        wrapper.setup(setup_dirs=True)
        wrapper.run()

def test_except_patterndoc():
    with wrap() as wrapper:
        with open("exceptme.abc", "w") as f:
            f.write("hello")

        parser = YamlFileParser(wrapper)
        parser.parse(""".abc:\n  - except : 'exceptme.abc' """)
        wrapper.batch.run()

        assert len(wrapper.batch.lookup_table) == 1

def test_except_patterndoc_pattern():
    with wrap() as wrapper:
        with open("exceptme.abc", "w") as f:
            f.write("hello")

        parser = YamlFileParser(wrapper)
        parser.parse(""".abc:\n  - except : 'exceptme.*' """)
        wrapper.batch.run()

        assert len(wrapper.batch.lookup_table) == 1

def test_children_siblings_order():
    with wrap() as wrapper:
        parser = YamlFileParser(wrapper)
        parser.parse("""
        p1:
            - c1
            - c2:
                - g1
                - g2
                - g3
            - c3
        """)

        wrapper.batch.run()

        p1 = wrapper.batch.lookup_table['BundleNode:p1']
        assert [i.key_with_class() for i in p1.walk_inputs()] == [
                'BundleNode:c1',
                'BundleNode:g1',
                'BundleNode:g2',
                'BundleNode:g3',
                'BundleNode:c2',
                'BundleNode:c3'
                ]

        c1 = wrapper.batch.lookup_table['BundleNode:c1']
        assert len(c1.inputs) == 0

        c2 = wrapper.batch.lookup_table['BundleNode:c2']
        assert [i.key_with_class() for i in c2.walk_inputs()] == [
                'BundleNode:g1',
                'BundleNode:g2',
                'BundleNode:g3'
                ]

        c3 = wrapper.batch.lookup_table['BundleNode:c3']
        assert len(c3.inputs) == 0

        g3 = wrapper.batch.lookup_table['BundleNode:g3']
        assert len(g3.inputs) == 0

def test_single_file_doc():
    with wrap() as wrapper:
        with open("hello.txt", "w") as f:
            f.write("hello")

        parser = YamlFileParser(wrapper)
        parser.parse("hello.txt")

        wrapper.batch.run()
        assert "Doc:hello.txt" in wrapper.batch.lookup_table

def test_single_bundle_doc():
    with wrap() as wrapper:
        parser = YamlFileParser(wrapper)
        parser.parse("hello")

        wrapper.batch.run()
        assert "BundleNode:hello" in wrapper.batch.lookup_table

def test_single_bundle_doc_with_args():
    with wrap() as wrapper:
        parser = YamlFileParser(wrapper)
        parser.parse("""
        more:
            - hello
            - one-more-task
            - foo: bar

        hello:
            - foo: bar
            - filter_fruit: orange
            - args:
                - ping: pong
            - another-task:
                - foo: baz
                - yet-another-task:
                    - foo: bar
            - one-more-task
        """)

        wrapper.batch.run()

        assert wrapper.batch.tree[0].key_with_class() == "BundleNode:more"
        assert len(wrapper.batch.lookup_table) == 5

def test_single_bundle_doc_with_args_2():
    with wrap() as wrapper:
        parser = YamlFileParser(wrapper)
        parser.parse("""

      -  hello:
            - foo: bar
            - filter_fruit: orange
            - args:
                - ping: pong
            - another-task:
                - foo: baz
                - yet-another-task:
                    - foo: bar
            - one-more-task

      -  more:
            - hello
            - one-more-task
            - foo: bar

        """)

        wrapper.batch.run()

        assert wrapper.batch.tree[0].key_with_class() == "BundleNode:more"
        assert len(wrapper.batch.lookup_table) == 5



from dexy.common import OrderedDict
from dexy.filter import DexyFilter
import os
import re

class HtmlSectionsFilter(DexyFilter):
    """
    Split content according to HTML comments.
    """
    ALIASES = ['htmlsections']
    OUTPUT_EXTENSIONS = ['.json']
    OUTPUT_DATA_TYPE = 'sectioned'
    FRAGMENT = False

    def process(self):
        output = OrderedDict()

        sections = re.split("<!-- section \"(.+)\" -->\n", self.input().as_text())

        for i in range(1, len(sections), 2):
            section_name = sections[i]
            section_content = sections[i+1]
            output[section_name] = section_content

        self.output().set_data(output)

class SplitHtmlFilter(DexyFilter):
    """
    Create multiple HTML pages from a single template, with an automatic index page.

    The split filter looks for specially formatted HTML comments in your
    document and splits your HTML into separate pages at each split comment.
    """
    ALIASES = ['split', 'splithtml']
    INPUT_EXTENSIONS = [".html"]
    OUTPUT_EXTENSIONS = [".html"]
    FRAGMENT = False

    def process(self):
        input_text = self.input().data()

        if input_text.find("<!-- endsplit -->") > 0:
            body, footer = re.split("<!-- endsplit -->", input_text, maxsplit=1)
            sections = re.split("<!-- split \"(.+)\" -->\n", body)
            header = sections[0]

            pages = OrderedDict()
            index_content = None
            for i in range(1, len(sections), 2):
                if sections[i] == 'index':
                    index_content = sections[i+1]
                else:
                    section_label = sections[i]
                    section_url = section_label.split(" ")[0]

                    filename = "%s.html" % section_url
                    filepath = os.path.join(self.output().parent_dir(), filename)
                    pages[section_label] = filename

                    new_page = self.add_doc(filepath, header + sections[i+1] + footer)
                    new_page.args['title'] = section_label

                    self.artifact.log.debug("added key %s to artifact %s ; links to file %s" %
                              (filepath, self.artifact.key, new_page.name))

            index_items = []
            for k in sorted(pages.keys()):
                index_items.append("""<li><a href="%s">%s</a></li>""" %
                                   (pages[k], k))

            output = []
            output.append(header)
            if index_content:
                output.append(index_content)

            if self.artifact.args.has_key("split-ul-class"):
                ul = "<ul class=\"%s\">" % self.artifact.args['split-ul-class']
            else:
                ul = "<ul class=\"split\">"

            output.append("%s\n%s\n</ul>" % (ul, "\n".join(index_items)))
            output.append(footer)
        else:
            # No endsplit found, do nothing.
            output = input_text

        self.output().set_data("\n".join(output))



from dexy.node import PatternNode
from dexy.node import DocNode
from dexy.tests.utils import TEST_DATA_DIR
from dexy.tests.utils import wrap
import inspect
import os
import shutil

def test_ragel_state_chart_to_image():
    ragel = inspect.cleandoc("""
        %%{
          machine hello_and_welcome;
          main := ( 'h' @ { puts "hello world!" }
                  | 'w' @ { puts "welcome" }
                  )*;
        }%%
          data = 'whwwwwhw'
          %% write data;
          %% write init;
          %% write exec;
        """)
    with wrap() as wrapper:
        node = DocNode("example.rl|rlrbd|dot",
                contents=ragel,
                wrapper=wrapper)
        wrapper.run_docs(node)

def test_latex_filter_with_unicode():
    with wrap() as wrapper:
        for f in ["test-unicode-latex-jinja.tex", "test-idio.py"]:
            start = os.path.join(TEST_DATA_DIR, f)
            shutil.copyfile(start, f)

        doc = PatternNode("*.tex|jinja|latex",
                inputs = [
                    PatternNode("*.py|idio|pycon|pyg|l", wrapper=wrapper)
                    ],
                wrapper=wrapper)
        wrapper.run_docs(doc)



import StringIO
import dexy.artifact
import dexy.plugin
import logging
import os
import shutil

class Reporter(object):
    TITLE = None
    ALIASES = []
    __metaclass__ = dexy.plugin.PluginMeta

    @classmethod
    def is_active(klass):
        return True

    ALLREPORTS = True # whether to be included in 'allreports', set to false for reporters with side effects
    REPORTS_DIR = None
    SAFETY_FILENAME = ".dexy-generated"
    README_FILENAME = "README"

    @classmethod
    def create_reports_dir(self, reports_dir=None):
        if not reports_dir:
            reports_dir = self.REPORTS_DIR

        self.remove_reports_dir(reports_dir, keep_empty_dir=True)

        safety_filepath = os.path.join(reports_dir, self.SAFETY_FILENAME)
        readme_filepath = os.path.join(reports_dir, self.README_FILENAME)

        if not os.path.exists(reports_dir):
            os.mkdir(reports_dir)

        with open(safety_filepath, "w") as f:
            f.write("""
            This directory was generated by the %s Dexy Reporter and
            may be deleted without notice.\n\n""" % self.__class__.__name__)
        with open(readme_filepath, "w") as f:
            f.write("""
            This directory was generated by the %s Dexy Reporter and
            may be deleted without notice.\n\n""" % self.__class__.__name__)

    @classmethod
    def remove_reports_dir(self, reports_dir=None, keep_empty_dir=False):
        if not reports_dir:
            reports_dir = self.REPORTS_DIR

        if not reports_dir:
            return False

        safety_filepath = os.path.join(reports_dir, self.SAFETY_FILENAME)

        if os.path.exists(reports_dir) and not os.path.exists(safety_filepath):
            raise Exception("Please remove directory %s, Dexy wants to put a report here but doesn't want to overwrite anything by accident." % os.path.abspath(reports_dir))
        elif os.path.exists(reports_dir):
            if keep_empty_dir:
                # Does not remove the base directory, useful if you are running
                # a process (like 'dexy serve') from inside that directory
                for f in os.listdir(reports_dir):
                    path = os.path.join(reports_dir, f)
                    if os.path.isdir(path):
                        shutil.rmtree(path)
                    else:
                        os.remove(path)
            else:
                shutil.rmtree(reports_dir)

    def run(self, wrapper):
        pass

    def set_log(self):
        if not hasattr(self, 'log'):
            self.log = logging.getLogger("report:%s" % self.ALIASES[0])
            self.logstream = StringIO.StringIO()
            handler = logging.StreamHandler(self.logstream)
            if hasattr(self, 'wrapper'):
                handler.setFormatter(logging.Formatter(self.wrapper.log_format))
            self.log.addHandler(handler)
            self.log.setLevel(logging.DEBUG)

            try:
                self.log.addHandler(logging.getLogger('dexy').handlers[0])
            except IndexError:
                pass



from dexy.node import DocNode
from dexy.tests.utils import TEST_DATA_DIR
from dexy.tests.utils import assert_output
from dexy.tests.utils import runfilter
from dexy.tests.utils import wrap
from nose.exc import SkipTest
import os
import shutil

def test_phantomjs_render_filter():
    with runfilter("phrender", "<p>hello</p>") as doc:
        assert doc.output().is_cached()

def test_phantomjs_stdout_filter():
    assert_output('phantomjs', PHANTOM_JS, "Hello, world!\n")

def test_casperjs_svg2pdf_filter():
    # TODO find smaller file - make test go faster?
    with wrap() as wrapper:
        orig = os.path.join(TEST_DATA_DIR, 'butterfly.svg')
        shutil.copyfile(orig, 'butterfly.svg')
        node = DocNode("butterfly.svg|svg2pdf",
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().is_cached()
        assert doc.output().filesize() > 1000

def test_casperjs_stdout_filter():
    with wrap() as wrapper:
        node = DocNode("example.js|casperjs",
                contents=CASPER_JS,
                wrapper=wrapper)

        wrapper.run_docs(node)

        assert 'google.pdf' in wrapper.batch.doc_names()

        try:
            assert 'cookies.txt' in wrapper.batch.doc_names()
        except AssertionError as e:
            import urllib
            try:
                urllib.urlopen("http://google.com")
                raise e
            except IOError:
                raise SkipTest


PHANTOM_JS = """
console.log('Hello, world!');
phantom.exit();
"""

CASPER_JS = """
var links = [];
var casper = require('casper').create();

casper.start('http://google.com/', function() {
    this.capture('google.pdf');
});

casper.run();
"""
